from typing import List

from elements_api import ElementsAsyncClient
from elements_api.models.common_models_pb2 import Pagination
from elements_api.models.toi_pb2 import TOI, TOIDeleteRequest, TOIGetRequest, \
    TOICreateRequest, TOIListRequest


class APIToi:
    def __init__(self, client: ElementsAsyncClient, timeout):
        self.__timeout = timeout
        self.__client = client

    async def create(self, toi: TOI) -> TOI:
        """
        This will call the Create Time of Interest (toi) service and create a new TOI for usage in a computation.
        When defining a TOI please use an instance of a TOIBuilder

        :param toi: TOI - use TOIBuilder to build TOIs
        :return: (toi_id, toi_recurrence_ids)
        """

        # elements-api TOICreateRequest now expects the full TOI message nested under the `toi` field
        request = TOICreateRequest(toi=toi)

        response = await self.__client.api.toi.create(request, timeout=self.__timeout)
        return response.toi

    async def get(self, ids: List, page_size: int = 2) -> List[TOI]:
        """
        Retrieve the TOI for the specified toi_id. The user must have permission to access the specified TOI.

        :param ids:
        :param page_size:
        :return: List[TOI]
        """
        pagination = Pagination(
            page_token="1",
            page_size=page_size,
            next_page_token="2"
        )
        request = TOIGetRequest(
            ids=ids,
            # pagination=pagination
        )
        response = await self.__client.api.toi.get(request)
        return response.toi_objects

    async def list(self, **kwargs) -> List[TOI]:
        """
        Retrieves the TOI for all tois that the user has permissions for and that match the provided search filters.
        :param kwargs:
        :return: List[TOI]
        """
        message_fragments = []
        for kwarg in kwargs.keys():
            if 'search_text' == kwarg:
                message_fragments.append(TOIListRequest(
                    search_text=kwargs[kwarg]
                ))
        request = TOIListRequest()
        for fragment in message_fragments:
            request.MergeFrom(fragment)

        response = await self.__client.api.toi.list(request, timeout=self.__timeout)
        return response.toi_objects

    async def delete(self, ids: List):
        """
        Deletes a list of TOIs

        :param ids: list of toi_id uuids
        :return: EmptyResponse
        """

        request = TOIDeleteRequest(
            ids=ids
        )
        await self.__client.api.toi.delete_toi(request, timeout=self.__timeout)



# flake8: noqa: E501
import os
import re
import logging
import tempfile
from zipfile import ZipFile
from collections import defaultdict
from typing import AsyncIterator, Dict, List, Optional

import boto3
import httpx
from google.protobuf.timestamp_pb2 import Timestamp

from elements_api import ElementsAsyncClient
from elements_api.models.common_models_pb2 import Pagination
from elements_api.models.result_pb2 import ResultGetRequest, ResultGetResponse


class APIResult:
    def __init__(self, client: ElementsAsyncClient, timeout):
        self.__timeout = timeout
        self.__client = client
        # External HTTP endpoints (overridable via environment)
        self.__results_api_base_url = os.getenv(
            "RESULT_DATA_SOURCE_API_URL",
            "https://results.orbitalinsight.com"
        )
        self.__app_api_base_url = os.getenv(
            "TERRASCOPE_API_HOST"
        )
        self.__api_token = os.getenv("TERRASCOPE_API_TOKEN")
        self.__httpx_timeout = httpx.Timeout(
            float(self.__timeout), read=300.0
        )

    @staticmethod
    async def merge_download_files(
        algorithm_computation_id_to_data_type_to_downloaded_paths: 
            Dict[str, Dict[str, List[str]]],
        download_dir: str = None
    ) -> Dict[str, Dict[str, str]]:
        download_dir = os.getcwd() if not download_dir else download_dir
        if not os.path.exists(download_dir):
            os.makedirs(download_dir)

        algorithm_computation_id_to_data_type_to_merged_file =             defaultdict(lambda: defaultdict(str))
        for (algorithm_computation_id,
                data_type_to_downloaded_paths) in algorithm_computation_id_to_data_type_to_downloaded_paths.items():
            for data_type, downloaded_paths in data_type_to_downloaded_paths.items():
                merged_file_dir = os.path.join(download_dir, algorithm_computation_id)
                os.makedirs(merged_file_dir, exist_ok=True)
                merged_file = os.path.join(merged_file_dir, f'{data_type}.csv')
                logging.info(
                    f"Merging files for algorithm_computation_id {algorithm_computation_id} "
                    f"and data_type {data_type} to {merged_file}"
                )

                header_written = False
                with open(merged_file, 'w') as output_csv:
                    for idx, downloaded_path in enumerate(downloaded_paths):
                        with tempfile.TemporaryDirectory() as working_dir:
                            interim_path = f'{working_dir}/{idx}'
                            os.mkdir(interim_path)
                            with ZipFile(downloaded_path, 'r') as zip_ref:
                                zip_ref.extractall(interim_path)
                            csv_path = f'{interim_path}/{data_type}.csv'
                            with open(csv_path, 'r') as input_csv:
                                if header_written:
                                    next(input_csv)
                                output_csv.write(input_csv.read())
                                header_written = True
                algorithm_computation_id_to_data_type_to_merged_file[algorithm_computation_id][data_type] = merged_file
        return algorithm_computation_id_to_data_type_to_merged_file

    async def download(
        self, algorithm_computation_ids: List[str] = None,
        analysis_computation_ids: List[str] = None,
        source_aoi_version: int = None, dest_aoi_version: int = None,
        algo_config_class: str = None,
        algo_config_subclass: str = None, created_on: Timestamp = None,
        observation_start_ts: Timestamp = None,
        max_observation_start_ts: Timestamp = None, data_type: str = None,
        download_dir: str = None
    ) -> Dict[str, Dict[str, List[str]]]:
        """
        algorithm_computation_ids: [Required] List[str] - Algorithm computation IDs

        :return: Dict[str, Dict[str, List[str]]]: mapping of algorithm_computation_id_to_data_type_to_downloaded_paths
        """
        download_dir = os.getcwd() if not download_dir else download_dir

        result_get_responses = await self.get(
            algorithm_computation_ids=algorithm_computation_ids,
            analysis_computation_ids=analysis_computation_ids,
            source_aoi_version=source_aoi_version,
            dest_aoi_version=dest_aoi_version,
            algo_config_class=algo_config_class,
            algo_config_subclass=algo_config_subclass,
            created_on=created_on,
            observation_start_ts=observation_start_ts,
            max_observation_start_ts=max_observation_start_ts,
            include_export_files=True, data_type=data_type
        )

        result_export_credentials = result_get_responses[0].export_credentials
        credentials = result_export_credentials.credentials
        s3 = boto3.client(
            's3',
            aws_access_key_id=credentials.fields['AccessKeyId'].string_value,
            aws_session_token=credentials.fields['SessionToken'].string_value,
            aws_secret_access_key=credentials.fields['SecretAccessKey'].string_value
        )
        pattern = r"https://(.*?)\.s3"
        container_name = re.search(pattern, result_export_credentials.base_url_template).group(1)
        algorithm_computation_id_to_data_type_to_downloaded_paths =             defaultdict(lambda: defaultdict(list))

        for result_get_response in result_get_responses:
            for result in result_get_response.results:
                data_type = result.data_type
                algorithm_computation_id = result.algorithm_computation_id
                for observation in result.observations:
                    key_path = observation.export_file.url
                    if not key_path:
                        continue
                    full_download_path = download_dir + os.path.split(key_path)[0]
                    filename = 'results.zip'
                    os.makedirs(full_download_path)
                    downloaded_path = os.path.join(full_download_path, filename)
                    s3.download_file(container_name, key_path[1:], downloaded_path)
                    algorithm_computation_id_to_data_type_to_downloaded_paths[
                        algorithm_computation_id][data_type].append(downloaded_path)

        logging.info(
            f"Downloaded results for algorithm_computation_ids and data_types: "
            f"{algorithm_computation_id_to_data_type_to_downloaded_paths}"
        )
        return algorithm_computation_id_to_data_type_to_downloaded_paths

    async def get(
        self, algorithm_computation_ids: List[str] = None,
        analysis_computation_ids: List[str] = None,
        source_aoi_version: int = None, dest_aoi_version: int = None,
        algo_config_class: str = None,
        algo_config_subclass: str = None, created_on: Timestamp = None,
        observation_start_ts: Timestamp = None,
        max_observation_start_ts: Timestamp = None,
        include_export_files: bool = None, data_type: str = None
    ) -> List[ResultGetResponse]:
        """
            required: algorithm_computation_ids or analysis_computation_ids
        """
        # Query all GetResultResponses
        result_get_responses = []
        pagination = Pagination(page_size=1000)
        has_next_result = True
        while has_next_result:
            request = ResultGetRequest(
                algorithm_computation_ids=algorithm_computation_ids,
                analysis_computation_ids=analysis_computation_ids,
                source_aoi_version=source_aoi_version,
                dest_aoi_version=dest_aoi_version,
                algo_config_class=algo_config_class,
                algo_config_subclass=algo_config_subclass,
                created_on=created_on,
                observation_start_ts=observation_start_ts,
                max_observation_start_ts=max_observation_start_ts,
                include_export_files=include_export_files,
                data_type=data_type,
                pagination=Pagination(
                    page_token=pagination.next_page_token, page_size=1000
                )
            )
            result_get_response = await self.__client.api.result.get(request, timeout=self.__timeout)
            result_get_responses.append(result_get_response)
            pagination = result_get_response.pagination
            has_next_result = pagination and pagination.next_page_token
        return result_get_responses

    # ---------------------------
    # Results Data Source methods
    # ---------------------------
    async def list_available_collections(self, project_id: str) -> List[str]:
        """
        List available collection IDs for a Terrascope project via the app API.

        Returns a list of strings formatted as "<algorithm_computation_id>---<data_type>".
        """
        if not self.__api_token:
            logging.error("TERRASCOPE_API_TOKEN not set; cannot list collections")
            return []

        url = f"{self.__app_api_base_url}/projects/{project_id}/summary"
        headers = {"Authorization": f"Bearer {self.__api_token}"}
        try:
            async with httpx.AsyncClient(timeout=self.__httpx_timeout) as client:
                r = await client.get(url, headers=headers)
                if r.status_code != 200:
                    logging.error(
                        "Failed to fetch project summary for %s: %s %s",
                        project_id,
                        r.status_code,
                        r.text,
                    )
                    return []
                summary_info = r.json()
        except Exception as e:
            logging.exception("Exception fetching project summary for %s: %s", project_id, e)
            return []

        collection_ids: List[str] = []
        analysis_info = summary_info.get(
            "data", {}
        ).get("analysis_computation_info", {})
        for _, computations in analysis_info.items():
            for computation_info in computations:
                computation_id = computation_info.get("algorithm_computation_id")
                output_types = computation_info.get(
                    "algorithm_config_details", {}
                ).get("output_data_types", [])
                for data_type in output_types:
                    collection_ids.append(f"{computation_id}---{data_type}")
        return collection_ids

    async def iter_arrow_batches(self, collection_id: str) -> AsyncIterator[object]:
        """
        Stream Arrow RecordBatches for a given collection_id from the Results Data Source API.
        """
        try:
            import pyarrow.ipc as ipc  # type: ignore
        except ImportError as e:
            raise ImportError(
                "pyarrow is required for iter_arrow_batches; install elements-sdk with pyarrow."
            ) from e

        if not self.__api_token:
            raise ValueError("TERRASCOPE_API_TOKEN not set; cannot fetch arrow stream")

        headers = {"Authorization": f"Bearer {self.__api_token}"}
        feature_url = f"{self.__results_api_base_url}/features/{collection_id}"

        async with httpx.AsyncClient(timeout=self.__httpx_timeout) as client:
            try:
                async with client.stream("GET", feature_url, headers=headers) as response:
                    if response.status_code != 200:
                        body = await response.aread()
                        logging.error(
                            "Failed to get Arrow stream for %s: %s, body: %s",
                            collection_id,
                            response.status_code,
                            body.decode(errors="ignore"),
                        )
                        return
                    with tempfile.SpooledTemporaryFile() as tmp:
                        async for chunk in response.aiter_bytes():
                            tmp.write(chunk)
                        if tmp.tell() == 0:
                            return
                        tmp.seek(0)
                        try:
                            with ipc.open_stream(tmp) as reader:
                                for batch in reader:
                                    yield batch
                        except Exception:
                            logging.exception("Failed to parse Arrow stream for %s", collection_id)
                            return
            except Exception:
                logging.exception("Exception fetching Arrow stream for %s", collection_id)
                return

    async def download_parquet(self, collection_id: str,
                                 dest_path: str) -> int:
        """
        Download a collection as Parquet to dest_path. Returns total rows written.
        """
        try:
            import pyarrow.parquet as pq  # type: ignore
        except ImportError as e:
            raise ImportError(
                "pyarrow is required for download_parquet; install elements-sdk with pyarrow."
            ) from e

        total_rows = 0
        writer = None
        try:
            async for batch in self.iter_arrow_batches(collection_id):
                if batch is None:
                    continue
                if writer is None:
                    writer = pq.ParquetWriter(dest_path, batch.schema)  # type: ignore
                writer.write_batch(batch)
                total_rows += batch.num_rows
        finally:
            try:
                if writer is not None:
                    writer.close()  # type: ignore
            except Exception:
                pass
        return total_rows

    async def get_collection_info(self, collection_id: str) -> Optional[dict]:
        """
        Retrieve collection info metadata from the Results Data Source API.
        """
        if not self.__api_token:
            logging.error("TERRASCOPE_API_TOKEN not set; cannot get collection info")
            return None
        headers = {"Authorization": f"Bearer {self.__api_token}"}
        url = f"{self.__results_api_base_url}/collections/{collection_id}"
        try:
            async with httpx.AsyncClient(timeout=self.__httpx_timeout) as client:
                response = await client.get(url, headers=headers)
                if response.status_code != 200:
                    logging.error(
                        "Failed to get collection info for %s: %s %s",
                        collection_id,
                        response.status_code,
                        await response.aread(),
                    )
                    return None
                return response.json()
        except Exception:
            logging.exception("Exception fetching collection info for %s", collection_id)
            return None



from typing import List

from elements_api import ElementsAsyncClient
from elements_api.models.permission_pb2 import (
    Permission, Resource, Subject, PermissionCreateRequest, PermissionCreateResponse,
    PermissionDeleteRequest, PermissionDeleteResponse
)
from elements_api.models.user_pb2 import User, UserGetRequest, UserListRequest


class APIPermission:
    def __init__(self, client: ElementsAsyncClient, timeout):
        self.__timeout = timeout
        self.__client = client

    async def create(self, resource_ids: List[str], resource_type: Resource.Type,
                     permission_type: Permission.Type, user_ids: List[str] = [],
                     user_emails: List[str] = [],
                     public: bool = False, public_confirm: bool = False) -> PermissionCreateResponse:
        """
        :param public:
        :param resource_ids: The resource IDs you'd like to share with the users.
        :param user_ids: The unique user IDs you'd like to share the resources with.
        :param user_emails: The user emails you'd like to share the resources with.
        :param resource_type: The type of resource the resource IDs reference.
        :param permission_type: Permission type to grant the users for the resource (read, write, admin).
        :param public: Flag to share publicly (all users). Must have global admin role and use the public_confirm param.
        :param public_confirm: Flag to confirm the sharing of the resource_ids publicly.
        :return: PermissionCreateResponse
        """
        # include users specified by uuid + user_ids
        uuids = await self.get_user_uuids(user_ids, user_emails, public, public_confirm)

        #  print(f'uuids= {uuids}')
        request = PermissionCreateRequest(
            permissions=[
                Permission(
                    resources=[
                        Resource(id=resource_id, type=resource_type)
                        for resource_id in resource_ids
                    ],
                    subjects=[
                        Subject(id=subject_uuid)
                        for subject_uuid in uuids
                    ],
                    permission_types=[permission_type]
                )
            ]
        )
        response = await self.__client.api.permission.create(request)
        return response

    async def get(self):
        """
        """
        pass

    async def get_user_uuids(self, user_ids: List[str] = [], user_emails: List[str] = [],
                             public: bool = False, public_confirm: bool = False) -> List:
        """
        :param public: Whether or not these are public resources
        :param user_ids: The unique user IDs you'd like to un-share the resources with.
        :param user_emails: The user emails you'd like to un-share the resources with.
        :param public: Flag to share publicly (all users). Must have global admin role and use the public_confirm param.
        :param public_confirm: Flag to confirm the removal of the shared resources publicly.
        :return: List
        """

        if public:
            if public_confirm:
                uuids = ["1"]
            else:
                print("You must set public_confirm=True to share or remove the public sharing of these resources")
                return
        else:
            # Get UUID of user from emails
            uuids = []

            for user_id in user_ids:
                uuids.append(user_id)

            for user_email in user_emails:
                user_list_request = UserListRequest(email=user_email)
                user_list_response = await self.__client.api.user.list(user_list_request)

                uuids += [user.id for user in user_list_response.users]

        # include users specified by uuid + user_ids
        return list(set(uuids))

    async def delete(self, resource_ids: List[str], resource_type: Resource.Type,
                     permission_type: Permission.Type, user_ids: List[str] = [],
                     user_emails: List[str] = [],
                     public: bool = False, public_confirm: bool = False) -> PermissionDeleteResponse:
        """
        :param public: Whether or not these are public resources
        :param resource_ids: The resource IDs you'd like to un-share with the users.
        :param user_ids: The unique user IDs you'd like to un-share the resources with.
        :param user_emails: The user emails you'd like to un-share the resources with.
        :param resource_type: The type of resource the resource IDs reference.
        :param permission_type: Permission type to grant the users for the resource (read, write, admin).
        :param public: Flag to share publicly (all users). Must have global admin role and use the public_confirm param.
        :param public_confirm: Flag to confirm the removal of the shared resources publicly.
        :return: PermissionDeleteResponse
        """
        uuids = await self.get_user_uuids(user_ids, user_emails, public, public_confirm)

        request = PermissionDeleteRequest(
            permissions=[
                Permission(
                    resources=[
                        Resource(id=resource_id, type=resource_type)
                        for resource_id in resource_ids
                    ],
                    subjects=[
                        Subject(id=subject_uuid)
                        for subject_uuid in uuids
                    ],
                    permission_types=[permission_type]
                )
            ]
        )
        response = await self.__client.api.permission.delete(request)
        return response



#!/usr/bin/env python

import click
import asyncio
import yaml
import json
import elements.cli.lib.workflow as wf
import elements.cli.lib.utils as tsu
from elements.cli.lib.aliased_group import AliasedGroup


@click.command(cls=AliasedGroup, help="'tasks' command group")
@click.pass_context
def tasks(ctx):
    pass


@tasks.command('initialize')
@click.pass_context
@click.option('-m', '--manifest_yaml', type=str)
@click.option('-om', '--output_manifest', type=str, help="Can be json or yaml file")
@click.option('-na', '--algorithm_name', type=str)
@click.option('-a', '--author', type=str, default='TERRASCOPE_AUTHOR')
@click.option('-N', '--display_name', type=str)
@click.option('-d', '--docker_version_hash', type=str)
@click.option('-pv', '--value_price', default=wf.VALUE_PRICE_DEFAULT, type=float)
@click.option('-pe', '--execution_price', default=wf.EXECUTION_PRICE_DEFAULT, type=float)
@click.option('-nc', '--algorithm_config_name', default=None, type=str)
@click.option('-nC', '--analysis_config_name', type=str, default=None,
              help="Name used in TS (default to 'display_name' from algo manifest)")
@click.option('-dC', '--analysis_config_desc', type=str, default=None,
              help="Description used in TS (default to 'description' from algo manifest)")
@click.option('-nA', '--analysis_name', default=None, type=str,
              help="Analysis name (defaults to algorithm_name if not provided).")
@click.option('-dA', '--analysis_description', default=None, type=str,
              help="Analysis description (default to 'description' from algo manifest).")
@click.option('--dry_run', is_flag=True, default=False)
@click.option('-V', '--version', type=str, default=None)
@click.option('-v', '--verbose', is_flag=True, default=False)
def tasks_initialize(
        ctx,
        manifest_yaml,
        output_manifest=None,
        algorithm_name=None,
        author=None,
        display_name=None,
        docker_version_hash=None,
        value_price=wf.VALUE_PRICE_DEFAULT,
        execution_price=wf.EXECUTION_PRICE_DEFAULT,
        algorithm_config_name=None,
        analysis_config_name=None,
        analysis_config_desc=None,
        analysis_name=None,
        analysis_description=None,
        dry_run=False,
        version=None,
        verbose=False,
):

    wf.check_environment_complete(raise_on_failure=True)

    # create the manifest
    input_manifest = None
    with open(manifest_yaml, 'r') as fp:
        input_manifest = yaml.safe_load(fp)
    algorithm_name = tsu.override_manifest_param('name', algorithm_name, input_manifest, do_overwrite=True)

    display_name = tsu.override_manifest_param('display_name', display_name, input_manifest, do_overwrite=True)
    author = tsu.get_author(author, raise_on_failure=True)
    visualizer_config_names = input_manifest.get('visualizer_config_names', None)

    # if the author is in the manifest, overwrite it with the command line / env-var value.
    # This is only done in case the user wants to write the manifest to a file,
    # ... we want that file to contain what was actually used.
    if 'author' in input_manifest:
        input_manifest['author'] = author

    image = wf.update_manifest_docker_hash(input_manifest, docker_version_hash)
    manifest = wf.create_algorithm_manifest(input_manifest)

    click.echo("Using the following information:")
    click.echo(f"--> name:         {algorithm_name}")
    click.echo(f"--> author:       {author}")
    click.echo(f"--> display_name: {display_name}")
    click.echo(f"--> image:        {image}")
    click.echo(f"--> prices:       {value_price} / {execution_price}")
    click.echo(f"--> viz-names:    {visualizer_config_names}")

    ############################
    # algorithm and algo version
    ############################
    if not dry_run:

        algorithm_id, algorithm_version_id = asyncio.run(wf.new_algorithm(
            algorithm_name,
            author,
            display_name,
            manifest,
            value_price=value_price,
            execution_price=execution_price,
            visualizer_config_names=visualizer_config_names,
        ))
        tsu.echo_highlight_suffix("You're new algorithm_id is: ", algorithm_id, 'green')
        tsu.echo_highlight_suffix("You're new algorithm_version_id is: ", algorithm_version_id, 'green')

    ############################
    # algo config
    ############################
    if not dry_run:

        if not algorithm_config_name:
            algorithm_config_name = algorithm_name
        config_desc = f"Initial algorithm config for {algorithm_config_name}"

        # Get the necessary params from the manifest.
        inputs = input_manifest['inputs'][0]

        try:
            data_type = inputs['data_type_name']
        except Exception:
            click.secho("Data Type must be provided in manifest", fg='red')
            raise

        # Try to find the data source
        data_source = inputs.get('data_source_name', None)
        if data_source is None:
            data_source = inputs.get('data_source_id', None)
        if data_source is None:
            data_source = inputs.get('data_source_ids', [None])[0]

        other_inputs = {}
        for key in ['min_count', 'max_count', 'cloud_paths']:
            if key in inputs:
                other_inputs[key] = inputs[key]

        image_processing_spec = None
        if 'parameters' in inputs:
            image_processing_spec = inputs['parameters'].get('image_processing_spec', None)
        data_parameters = {}
        if image_processing_spec:
            data_parameters['image_processing_spec'] = image_processing_spec

        algorithm_config_id = asyncio.run(
            wf.create_algorithm_config(
                algorithm_version_id,
                algorithm_config_name,
                config_desc,
                tsu.FakeDataSource(data_source),
                data_type,
                data_parameters,
                visualizer_config_names=visualizer_config_names,
                **other_inputs
            )
        )
        tsu.echo_highlight_suffix("You're new algorithm_config_id is: ", algorithm_config_id, 'green')

    ############################
    # analysis and analysis version
    ############################

    if not analysis_name:
        analysis_name = algorithm_name
    if not analysis_description:
        analysis_description = input_manifest['metadata']['description']
    tags = input_manifest['metadata']['tags']

    # create the analysis
    if not dry_run:
        analysis_id = asyncio.run(wf.create_analysis(
            analysis_name,
            author,
        ))
        tsu.echo_highlight_suffix("You're new analysis_id is: ", analysis_id, 'green')

        # create a manifest
        manifest = wf.create_analysis_manifest(analysis_name,
                                               algorithm_version_id,
                                               analysis_description,
                                               tags)
        # create the analysis version
        analysis_version_id = asyncio.run(wf.create_analysis_version(
            analysis_id,
            manifest,
        ))

        tsu.echo_highlight_suffix("You're new analysis_version_id is: ", analysis_version_id, 'green')

    ############################
    # analysis config
    ############################
    if not analysis_config_name:
        analysis_config_name = analysis_name
    if not analysis_config_desc:
        analysis_config_desc = analysis_description

    click.echo("Using the following information:")
    click.echo(f"--> analysis_version_id:      {analysis_version_id}")
    click.echo(f"--> algorithm_config_id:      {algorithm_config_id}")
    click.echo(f"--> analysis_config_name:     {analysis_config_name}")
    click.echo(f"--> analysis_config_desc:     {analysis_config_desc}")

    if not dry_run:

        analysis_config_id = asyncio.run(
            wf.create_analysis_config(
                analysis_version_id,
                algorithm_config_id,
                analysis_config_name,
                analysis_config_desc,
            )
        )
        tsu.echo_highlight_suffix("You're new analysis_config_id is: ", analysis_config_id, 'green')

    if output_manifest:
        with open(output_manifest, 'w') as fp:
            if output_manifest.endswith('json'):
                json.dump(input_manifest, fp, indent=4)
            else:
                yaml.dump(input_manifest, fp)


@tasks.command('update')
@click.pass_context
@click.option('-ia', '--algorithm_id', type=str, help="Algorithm ID to update")
@click.option('-iA', '--analysis_id', type=str, help="Analysis ID to update")
@click.option('-m', '--manifest_yaml', type=str)
@click.option('-om', '--output_manifest', type=str, help="Can be json or yaml file")
@click.option('-na', '--algorithm_name', type=str)
@click.option('-a', '--author', type=str, default='TERRASCOPE_AUTHOR')
@click.option('-N', '--display_name', type=str)
@click.option('-d', '--docker_version_hash', type=str)
@click.option('-pv', '--value_price', default=wf.VALUE_PRICE_DEFAULT, type=float)
@click.option('-pe', '--execution_price', default=wf.EXECUTION_PRICE_DEFAULT, type=float)
@click.option('-nc', '--algorithm_config_name', type=str, default=None)
@click.option('-nC', '--analysis_config_name', type=str, default=None,
              help="Name used in TS (default to 'display_name' from algo manifest)")
@click.option('-dC', '--analysis_config_desc', type=str, default=None,
              help="Description used in TS (default to 'description' from algo manifest)")
@click.option('-nA', '--analysis_name', default=None, type=str,
              help="Analysis name (defaults to algorithm_name if not provided).")
@click.option('-dA', '--analysis_description', default=None, type=str,
              help="Description which appears in TS (default to 'description' from algo manifest).")
@click.option('--dry_run', is_flag=True, default=False)
@click.option('-v', '--verbose', is_flag=True, default=False)
def tasks_update(
        ctx,
        algorithm_id,
        analysis_id,
        manifest_yaml,
        output_manifest=None,
        algorithm_name=None,
        author=None,
        display_name=None,
        docker_version_hash=None,
        value_price=wf.VALUE_PRICE_DEFAULT,
        execution_price=wf.EXECUTION_PRICE_DEFAULT,
        algorithm_config_name=None,
        analysis_config_name=None,
        analysis_config_desc=None,
        analysis_name=None,
        analysis_description=None,
        dry_run=False,
        verbose=False,
):

    wf.check_environment_complete(raise_on_failure=True)

    # create the manifest
    input_manifest = None
    with open(manifest_yaml, 'r') as fp:
        input_manifest = yaml.safe_load(fp)
    algorithm_name = tsu.override_manifest_param('name', algorithm_name, input_manifest, do_overwrite=True)

    if algorithm_config_name is None:
        algorithm_config_name = algorithm_name

    display_name = tsu.override_manifest_param('display_name', display_name, input_manifest, do_overwrite=True)
    author = tsu.get_author(author, raise_on_failure=True)
    visualizer_config_names = input_manifest.get('visualizer_config_names', None)

    # if the author is in the manifest, overwrite it with the command line / env-var value.
    # This is only done in case the user wants to write the manifest to a file,
    # ... we want that file to contain what was actually used.
    if 'author' in input_manifest:
        input_manifest['author'] = author

    image = wf.update_manifest_docker_hash(input_manifest, docker_version_hash)
    manifest = wf.create_algorithm_manifest(input_manifest)

    click.echo("Using the following information:")
    click.echo(f"--> algorithm_id: {algorithm_id}")
    click.echo(f"--> analysis_id:  {analysis_id}")
    click.echo(f"--> image:        {image}")
    click.echo(f"--> prices:       {value_price} / {execution_price}")
    click.echo(f"--> viz-names:    {visualizer_config_names}")

    ############################
    # algorithm and algo version
    ############################
    if not dry_run:

        tsu.echo_highlight_suffix("Updating existing algorithm_id: ", algorithm_id, 'green')

        algorithm_version_id = asyncio.run(wf.update_algorithm(
            algorithm_id,
            manifest,
            value_price=value_price,
            execution_price=execution_price,
            visualizer_config_names=visualizer_config_names,
        ))
        tsu.echo_highlight_suffix("Your updated algorithm_version_id is: ", algorithm_version_id, 'green')

    ############################
    # algo config
    ############################
    if not dry_run:

        if not algorithm_config_name:
            algorithm_config_name = algorithm_name
        config_desc = f"Initial algorithm config for {algorithm_config_name}"

        # Get the necessary params from the manifest.
        inputs = input_manifest['inputs'][0]

        try:
            data_type = inputs['data_type_name']
        except Exception:
            click.secho("Data Type must be provided in manifest", fg='red')
            raise

        data_source = inputs.get('data_source_name', None)
        if data_source is None:
            data_source = inputs.get('data_source_ids', [None])[0]

        other_inputs = {}
        for key in ['min_count', 'max_count', 'cloud_paths']:
            if key in inputs:
                other_inputs[key] = inputs[key]

        image_processing_spec = None
        if 'parameters' in inputs:
            image_processing_spec = inputs['parameters'].get('image_processing_spec', None)
        data_parameters = {}
        if image_processing_spec:
            data_parameters['image_processing_spec'] = image_processing_spec

        algorithm_config_id = asyncio.run(
            wf.create_algorithm_config(
                algorithm_version_id,
                algorithm_config_name,
                config_desc,
                tsu.FakeDataSource(data_source),
                data_type,
                data_parameters,
                visualizer_config_names=visualizer_config_names,
                **other_inputs
            )
        )
        tsu.echo_highlight_suffix("You're new algorithm_config_id is: ", algorithm_config_id, 'green')

    ############################
    # analysis version
    ############################

    if not analysis_name:
        analysis_name = algorithm_name
    if not analysis_description:
        analysis_description = input_manifest['metadata']['description']
    tags = input_manifest['metadata']['tags']

    # create the analysis
    if not dry_run:
        tsu.echo_highlight_suffix("Updating existing analysis_id: ", analysis_id, 'green')

        # create a manifest
        manifest = wf.create_analysis_manifest(analysis_name,
                                               algorithm_version_id,
                                               analysis_description,
                                               tags)
        # create the analysis version
        analysis_version_id = asyncio.run(wf.create_analysis_version(
            analysis_id,
            manifest,
        ))

        tsu.echo_highlight_suffix("You're new analysis_version_id is: ", analysis_version_id, 'green')

    ############################
    # analysis config
    ############################
    if not analysis_config_name:
        analysis_config_name = analysis_name
    if not analysis_config_desc:
        analysis_config_desc = analysis_description

    click.echo("Using the following information:")
    click.echo(f"--> analysis_version_id:      {analysis_version_id}")
    click.echo(f"--> algorithm_config_id:      {algorithm_config_id}")
    click.echo(f"--> analysis_config_desc:     {analysis_config_desc}")

    if not dry_run:

        analysis_config_id = asyncio.run(
            wf.create_analysis_config(
                analysis_version_id,
                algorithm_config_id,
                analysis_config_name,
                analysis_config_desc,
            )
        )
        tsu.echo_highlight_suffix("You're new analysis_config_id is: ", analysis_config_id, 'green')

    if output_manifest:
        with open(output_manifest, 'w') as fp:
            if output_manifest.endswith('json'):
                json.dump(input_manifest, fp, indent=4)
            else:
                yaml.dump(input_manifest, fp)


@tasks.command('idtrace')
@click.pass_context
@click.option('-iA', '--analysis_id', type=str, help="Analysis ID to trace")
@click.option('-iV', '--analysis_version_id', type=str, help="Analysis version ID to trace")
@click.option('-nA', '--analysis_name', default=None, type=str,
              help="Analysis name (defaults to algorithm_name if not provided).")
@click.option('--dry_run', is_flag=True, default=False)
@click.option('-v', '--verbose', is_flag=True, default=False)
def tasks_idtrace(
        ctx,
        analysis_id,
        analysis_version_id,
        analysis_name=None,
        dry_run=False,
        verbose=False,
):

    wf.check_environment_complete(raise_on_failure=True)

    # if they provide the analysis_version_id, look up the analysis_id and go from there.
    if analysis_version_id and not analysis_id and not analysis_name:
        analysis_version = asyncio.run(wf.get_analysis_versions(analysis_version_id=analysis_version_id))[0]
        analysis_version_dict = tsu.protobuf_to_dict(analysis_version)
        analysis_id = analysis_version_dict.get('analysis.id', None)

    ###############################################
    # ana id - get versions, then configs
    # If they give us a uuid, use it.  Otherwise, assume it's a name and look it up.
    if analysis_id:
        analyses = asyncio.run(wf.get_analyses([analysis_id]))
    if analysis_name:
        analyses = asyncio.run(wf.list_analyses(analysis_name))
    analysis_ids = [[analysis.id, analysis.name, analysis.author] for analysis in analyses]

    click.secho(f"input analysis_id:   {analysis_id}", fg='cyan')
    click.secho(f"input analysis_name: {analysis_name}", fg='cyan')
    indent = '  '
    for analysis_id, name, author in analysis_ids:

        click.secho("{:30s}  {:36s}  {:30.30s}".format("name", "analysis_id", "author"), fg='cyan')
        click.echo(f"{name:<30s}  {analysis_id:36s}  {author:30.30s}")

        ###################################################
        # get the analysis_version_id
        analysis_versions = asyncio.run(wf.get_analysis_versions(analysis_id=analysis_id))

        for analysis_version in analysis_versions:
            analysis_version_dict = tsu.protobuf_to_dict(analysis_version)

            analysis_version_dict['created_on'] = analysis_version_dict['created_on'][:-8]
            columns = ['analysis_version_id', 'created_on', 'analysis.name']
            click.secho(f"{indent}" + "{:36s} {:20s} {:20s}".format(*columns), fg='yellow')
            columns[0] = 'id'
            click.echo(f"{indent}" + "{:36s} {:20s} {:20s}".format(*[analysis_version_dict[c] for c in columns]))

            # if the use provided an analysis_version_id, only show that one
            if analysis_version_id and analysis_version_dict['id'] != analysis_version_id:
                continue

            analysis_version_detail = asyncio.run(wf.get_analysis_versions(analysis_version_id=analysis_version_dict['id']))[0]
            analysis_version_detail_dict = tsu.protobuf_to_dict(analysis_version_detail)
            algorithm_versions = analysis_version_detail_dict.get('algorithm_versions', None)
            for algorithm_version in algorithm_versions:
                algorithm_id = algorithm_version.get('algorithm', {}).get('id', None)
                algorithm_name = algorithm_version.get('algorithm', {}).get('name', None)
                click.secho(f"{indent}" + "{:26s}".format("--> algorithm_id: "), fg='magenta', nl=False)
                click.echo(f"{algorithm_id} ({algorithm_name})")
                algorithm_version_id = algorithm_version.get('id', None)
                version = algorithm_version.get('version', None)
                click.secho(f"{indent}" + "{:26s}".format("--> algorithm_version_id: "), fg='magenta', nl=False)
                click.echo(f"{algorithm_version_id} ({version})")

            ###################################################
            # get the analysis_config_id
            analysis_configs = asyncio.run(wf.get_analysis_configs(analysis_version_id=analysis_version_dict['id']))

            for analysis_config in analysis_configs:
                ac_dict = tsu.protobuf_to_dict(analysis_config)
                # analysis_config_id = ac_dict['id']
                ac_dict['created_on'] = ac_dict['created_on'][:-8]
                columns = ['analysis_config_id', 'created_on', 'name']
                click.secho(f"{indent}{indent}" + "{:36s} {:28s} {:20s}".format(*columns), fg='yellow')
                columns[0] = 'id'
                click.echo(f"{indent}{indent}" + "{:36s} {:28s} ".format(*[ac_dict[c] for c in columns[:2]]), nl=False)
                click.secho("{:20s}".format(ac_dict['name']), fg='magenta')
                for node in ac_dict.get('algorithm_config_nodes', []):
                    algorithm_config_id = node['algorithm_config_id']
                    click.secho(f"{indent}{indent}" + "{:26s}".format("--> algorithm_config_id: "),
                                fg='magenta', nl=False)
                    click.echo(f"{algorithm_config_id} ({node['name']})")


tasks.add_command(tasks_update)
tasks.add_command(tasks_initialize)
tasks.add_command(tasks_idtrace)



#!/usr/bin/env python

import click
import asyncio
import pandas as pd
import elements.cli.lib.workflow as wf
import elements.cli.lib.utils as tsu
from elements.cli.lib.aliased_group import AliasedGroup
from elements_api.models.permission_pb2 import Permission


@click.command(cls=AliasedGroup, help="'permission' command group")
@click.pass_context
def permission(ctx):
    pass


@permission.command('get')
@click.pass_context
@click.option('-iC', '--analysis_config_id', type=str)
def permission_get(ctx, analysis_config_id):

    click.echo("I'm sorry.  permission 'get' command is not yet implemented")
    if 0:
        wf.check_environment_complete(raise_on_failure=True)
        permissions = asyncio.run(wf.get_permissions(analysis_config_id))
        data = []
        for permission in permissions:
            data.append(tsu.protobuf_to_dict(permission))
        df = pd.DataFrame(data)
        click.secho(f"analysis_config_id: {analysis_config_id}", fg='cyan')
        click.echo(df)


@permission.command('set')
@click.pass_context
@click.option('-iC', '--analysis_config_id', type=str)
@click.option('-u', '--user_emails', multiple=True, type=str)
@click.option('-p', '--public', is_flag=True, default=False)
# @click.option('-t', '--permission_type', type=click.Choice([Permission.Type.READ]),
#              default=Permission.Type.READ)
@click.option('--dry_run', is_flag=True)
def permission_set(ctx, analysis_config_id, user_emails, public, dry_run):

    permission_type = Permission.Type.READ

    wf.check_environment_complete(raise_on_failure=True)

    if not dry_run:
        # user_ids = asyncio.run(wf.get_userid_from_email(user_emails))
        asyncio.run(
            wf.analysis_permission_create(
                analysis_config_id,
                user_emails,
                permission_type=permission_type,
                public=public,
                public_confirm=public,
            )
        )



from typing import Optional

from shapely.geometry.base import BaseGeometry
from elements.sdk.tools.sdk_support import SDKSupport

from elements_api import ElementsAsyncClient
from elements_api.models.order_pb2 import Order, OrderCreateRequest, Item, OrderApproveRequest, OrderCancelRequest, \
    OrderGetRequest, OrderListRequest, OrderState


class APIOrder:
    def __init__(self, client: ElementsAsyncClient, timeout):
        self.__timeout = timeout
        self.__client = client
        self.sdk_support = SDKSupport()

    async def create(self, data_source_id: str, product_spec_name: str,
                     scene_id: str,
                     clip_geom: Optional[BaseGeometry] = None,
                     metadata: Optional[dict] = None) -> Order:
        request = OrderCreateRequest(
            data_source_id=data_source_id,
            product_spec_name=product_spec_name,
            metadata=metadata,
            items=[
                Item(scene_id=scene_id,
                     target_geom=clip_geom.wkb if clip_geom is not None else None)
            ]
        )
        response = await self.__client.api.order.create(request)
        return response.order

    async def approve(self, order_id: str, comment: str) -> None:
        request = OrderApproveRequest(
            id=order_id,
            comment=comment
        )
        await self.__client.api.order.approve(request)

    async def cancel(self, order_id: str) -> None:
        request = OrderCancelRequest(
            id=order_id
        )
        await self.__client.api.order.cancel(request)

    async def get(self, order_id: str) -> Order:
        request = OrderGetRequest(
            id=order_id
        )
        response = await self.__client.api.order.get(request)
        return response.order

    async def list(self,
                   data_source_ids: list[str],
                   states: Optional[list[str]] = None,
                   analysis_computation_ids: Optional[list[str]] = None,
                   algorithm_computation_ids: Optional[list[str]] = None) -> list[Order]:
        ord_states = [OrderState.Value(s) for s in states] if states else []
        request = OrderListRequest(
            data_source_ids=data_source_ids,
            states=ord_states,
            analysis_computation_ids=analysis_computation_ids,
            algorithm_computation_ids=algorithm_computation_ids
        )
        responses = await self.sdk_support.get_all_paginated_objects(request,
                                                                     api_function=self.__client.api.order.list,
                                                                     timeout=self.__timeout)
        orders = []
        for response in responses:
            orders.extend(response.orders)
        return orders



#!/usr/bin/env python

import click
import asyncio
import pandas as pd
import elements.cli.lib.workflow as wf
import elements.cli.lib.utils as tsu
from elements.cli.lib.aliased_group import AliasedGroup


@click.command(cls=AliasedGroup, help="'credit' command group")
@click.pass_context
def credit(ctx):
    pass


@credit.command('get')
@click.pass_context
@click.option('-iv', '--algorithm_version_id', type=str)
def credit_get(ctx, algorithm_version_id):

    wf.check_environment_complete(raise_on_failure=True)
    click.echo("I'm sorry.  credit 'get' command is not yet implemented")

    if 0:
        source_id = '12bb3434-4696-4796-9428-7b9886ef76da'
        algo_credits = asyncio.run(wf.get_credit(source_id))
        data = []
        for algo_credit in algo_credits:
            data.append(tsu.protobuf_to_dict(algo_credit))
        df = pd.DataFrame(data)
        click.secho(f"Credit source ID: {source_id}", fg='cyan')
        click.echo(df)


@credit.command('set')
@click.pass_context
@click.option('-iv', '--algorithm_version_id', type=str)
@click.option('-pv', '--value_price', type=float)
@click.option('-pe', '--execution_price', type=float)
@click.option('--dry_run', is_flag=True)
def credit_set(ctx, algorithm_version_id, value_price, execution_price, dry_run):

    wf.check_environment_complete(raise_on_failure=True)

    if not dry_run:
        asyncio.run(
            wf.set_credit(algorithm_version_id=algorithm_version_id,
                          value_price=value_price,
                          execution_price=execution_price)
        )




import click
import elements.cli.lib.workflow as wf
from elements.cli.lib.aliased_group import AliasedGroup


@click.command(cls=AliasedGroup, help="'environment' command group")
@click.pass_context
def environment(ctx):
    pass


@environment.command('check')
@click.pass_context
@click.option('-a', '--check_admin', is_flag=True, default=False)
@click.option('-v', '--verbose', is_flag=True, default=False)
def env_check(ctx, check_admin=False, verbose=False):

    is_complete = wf.check_environment_complete(raise_on_failure=False, print_missing=True, verbose=verbose,
                                                requires_admin=check_admin)
    knot = ''
    if is_complete:
        fg = 'green'
    else:
        fg = 'red'
        knot = 'NOT '
    click.secho(f"Environment is {knot}complete", fg=fg)



import logging
import time
import uuid
from datetime import datetime
import pytest

from elements_api.models.common_models_pb2 import Pagination
from elements.sdk.builder.algorithm import (
    AlgorithmManifest, InterfaceType, DataType, AlgorithmConfiguration, DataSource
)
from elements.sdk.builder.toi import Frequency
from elements.sdk.builder.toi import TOIBuilder, TOIRuleBuilder
from elements.sdk.elements_sdk import ElementsSDK


class TestAlgorithm:
    @pytest.mark.asyncio
    async def test_create(self):
        sdk = ElementsSDK()
        name = "device-visits-sdk-integration-test-{}".format(uuid.uuid4())
        display_name = "Device Visits SDK Integration"
        author = "elements.sdk.integration@orbitalinsight.com"

        algorithm = await sdk.algorithm.create(name=name,
                                               display_name=display_name,
                                               author=author)
        assert algorithm.id is not None
        assert algorithm.name == name
        assert algorithm.display_name == display_name
        assert algorithm.author == author

    @pytest.mark.asyncio
    @pytest.mark.parametrize("new_name, new_display_name, new_author", [
        ("RegularTestAlgorithm".format(uuid.uuid4()), "Regular Test Algorithm", "bang.test@bangtest.com")

    ])
    async def test_update(self,
                          new_name: str,
                          new_display_name: str,
                          new_author: str):
        sdk = ElementsSDK()
        name = "device-visits-sdk-integration-test-{}".format(uuid.uuid4())
        display_name = "Device Visits SDK Integration"
        author = "elements.sdk.integration@orbitalinsight.com"

        algorithm_model = await sdk.algorithm.create(name=name,
                                                     display_name=display_name,
                                                     author=author)
        assert algorithm_model.id is not None
        assert algorithm_model.name == name
        assert algorithm_model.display_name == display_name
        assert algorithm_model.author == author

        algorithm_model.display_name = new_display_name
        algorithm_model.author = new_author
        algorithms = [algorithm_model]
        await sdk.algorithm.update(algorithms=algorithms)
        algorithm_new_models = await sdk.algorithm.get([algorithm_model.id])
        assert len(algorithm_new_models) == 1
        algorithm_new_model = algorithm_new_models[0]
        assert algorithm_new_model.display_name == new_display_name
        assert algorithm_new_model.author == new_author

    @pytest.mark.asyncio
    @pytest.mark.parametrize("algorithm_count", [
        1,
        10
    ])
    async def test_get(self, algorithm_count):
        sdk = ElementsSDK()

        display_name = "Test Get SDK Integration"
        author = "elements.sdk.integration@orbitalinsight.com"

        algorithm_ids = []

        for i in range(algorithm_count):
            name = "sdk-integration-test-algorithm-get-{}".format(uuid.uuid4())
            algorithm_model = await sdk.algorithm.create(name=name,
                                                         display_name=display_name,
                                                         author=author)
            algorithm_ids.append(algorithm_model.id)

        algorithms = await sdk.algorithm.get(algorithm_ids, pagination=Pagination(page_size=5))
        for algorithm in algorithms:
            assert algorithm.display_name == display_name
            assert algorithm.author == author

        assert len(algorithm_ids) == len(algorithms)

    @pytest.mark.asyncio
    @pytest.mark.parametrize("algorithm_count", [
        10,
        20
    ])
    async def test_list(self, algorithm_count):
        sdk = ElementsSDK()
        search_key = uuid.uuid4()
        name = str(search_key)
        display_name = "sdk-integration-test-algorithm-get-{}".format(name)
        author = "elements.sdk.integration@orbitalinsight.com"

        algorithm_ids = []
        now = datetime.utcnow()
        halfway = None
        for i in range(algorithm_count):
            algorithm_model = await sdk.algorithm.create(name="{}-{}".format(name, i),
                                                         display_name=display_name,
                                                         author=author)
            if i + 1 == int(algorithm_count / 2):
                time.sleep(10)
                halfway = datetime.utcnow()
            algorithm_ids.append(algorithm_model.id)

        # search text
        algorithms = await sdk.algorithm.list(search_text=name)
        assert len(algorithm_ids) == len(algorithms)
        algorithms = await sdk.algorithm.list(min_created_on=now, max_created_on=halfway)
        assert len(algorithms) == algorithm_count / 2


class TestAlgorithmVersion:
    @pytest.mark.asyncio
    async def test_create(self):
        sdk = ElementsSDK()
        algo_1_name = "device-visits-sdk-integration-test-{}".format(uuid.uuid4())
        algo_1_display_name = "Device Visits SDK Integration"
        algo_1_author = "elements.sdk.integration@orbitalinsight.com"

        algorithm = await sdk.algorithm.create(name=algo_1_name,
                                               display_name=algo_1_display_name,
                                               author=algo_1_author)
        assert algorithm.id is not None
        assert algorithm.name == algo_1_name
        assert algorithm.display_name == algo_1_display_name
        assert algorithm.author == algo_1_author

        manifest = AlgorithmManifest()
        manifest.metadata_required(description="Testing algo manifest builder")
        manifest.interface_required(interface_type=InterfaceType.FILESYSTEM_TASK_WORKER.value)
        manifest.inputs_add_data_type(data_type_name=DataType.PINGS.value,
                                      min_count=2,
                                      max_count=2)

        manifest.container_parameters_required(image="orbitalinsight/raw_foottraffic:84c76f7f",
                                               command=["python",
                                                        "/orbital/base/algorithms/"
                                                        "raw_foottraffic/src/py/raw_foottraffic/simple_foottraffic.py"])
        manifest.outputs_add_data_type(data_type_name="device_visits",
                                       observation_value_columns=["unique_device_count"])

        algorithm_version = await sdk.algorithm_version.create(algorithm_id=algorithm.id,
                                                               algorithm_manifest=manifest)

        assert algorithm_version.id is not None

    @pytest.mark.asyncio
    @pytest.mark.parametrize("version_count", [
        1,
        10,
        20,
        45
    ])
    async def test_get(self, version_count: int):
        sdk = ElementsSDK()
        name = "device-visits-sdk-integration-test-{}".format(uuid.uuid4())
        display_name = "Device Visits SDK Integration"
        author = "elements.sdk.integration@orbitalinsight.com"

        algorithm = await sdk.algorithm.create(name=name,
                                               display_name=display_name,
                                               author=author)
        assert algorithm.id is not None
        assert algorithm.name == name
        assert algorithm.display_name == display_name
        assert algorithm.author == author

        algorithm_version_ids = []
        for i in range(version_count):
            manifest = AlgorithmManifest()
            manifest.metadata_required(description="Testing algo manifest builder")
            manifest.interface_required(interface_type=InterfaceType.FILESYSTEM_TASK_WORKER.value)
            manifest.inputs_add_data_type(data_type_name=DataType.PINGS.value,
                                          min_count=2,
                                          max_count=2)

            manifest.container_parameters_required(image="orbitalinsight/raw_foottraffic:84c76f7f",
                                                   command=["python",
                                                            "/orbital/base/algorithms/"
                                                            "raw_foottraffic/src/py/raw_foottraffic/simple_foottraffic.py"])
            manifest.outputs_add_data_type(data_type_name="device_visits",
                                           observation_value_columns=["unique_device_count"])

            algorithm_version = await sdk.algorithm_version.create(algorithm_id=algorithm.id,
                                                                   algorithm_manifest=manifest)

            assert algorithm_version.id is not None
            algorithm_version_ids.append(algorithm_version.id)
        algorithm_versions = await sdk.algorithm_version.get(algorithm_version_ids=algorithm_version_ids,
                                                             pagination=Pagination(page_size=5))
        assert len(algorithm_versions) == version_count

    @pytest.mark.asyncio
    @pytest.mark.parametrize("version_count", [
        10,
        20
    ])
    async def test_list(self, version_count):
        sdk = ElementsSDK()
        unique = uuid.uuid4()
        name = "device-visits-sdk-integration-test-{}".format(unique)
        display_name = str(unique)
        author = "elements.sdk.integration@orbitalinsight.com"

        algorithm = await sdk.algorithm.create(name=name,
                                               display_name=display_name,
                                               author=author)
        assert algorithm.id is not None
        assert algorithm.name == name
        assert algorithm.display_name == display_name
        assert algorithm.author == author
        tag_search_key = str(uuid.uuid4())
        algorithm_version_ids = []
        now = datetime.utcnow()
        halfway = None
        for i in range(version_count):
            manifest = AlgorithmManifest()
            manifest.metadata_required(
                description="Testing algo manifest builder")
            manifest.interface_required(interface_type=InterfaceType.FILESYSTEM_TASK_WORKER.value)
            manifest.inputs_add_data_type(data_type_name=DataType.PINGS.value,
                                          min_count=2,
                                          max_count=2)

            manifest.container_parameters_required(image="orbitalinsight/raw_foottraffic:84c76f7f",
                                                   command=["python",
                                                            "/orbital/base/algorithms/"
                                                            "raw_foottraffic/src/py/raw_foottraffic/simple_foottraffic.py"])
            manifest.outputs_add_data_type(data_type_name="device_visits",
                                           observation_value_columns=["unique_device_count"])
            manifest.metadata_tags([tag_search_key])

            algorithm_version = await sdk.algorithm_version.create(algorithm_id=algorithm.id,
                                                                   algorithm_manifest=manifest)
            if i + 1 == version_count / 2:
                halfway = datetime.utcnow()
            assert algorithm_version.id is not None
            algorithm_version_ids.append(algorithm_version.id)

        algorithm_versions = await sdk.algorithm_version.list(algorithm_id=algorithm.id,
                                                              include_all_versions=True)
        assert len(algorithm_versions) == len(algorithm_version_ids)

        algorithm_versions = await sdk.algorithm_version.list(tag=tag_search_key, include_all_versions=True)
        assert len(algorithm_versions) == version_count
        algorithm_versions = await sdk.algorithm_version.list(min_created_on=now,
                                                              max_created_on=halfway, include_all_versions=True)
        assert len(algorithm_versions) == version_count / 2

    @pytest.mark.asyncio
    async def test_deprecate(self):
        sdk = ElementsSDK()
        name = "device-visits-sdk-integration-test-{}".format(uuid.uuid4())
        display_name = "Device Visits SDK Integration"
        author = "elements.sdk.integration@orbitalinsight.com"

        algorithm = await sdk.algorithm.create(name=name,
                                               display_name=display_name,
                                               author=author)
        assert algorithm.id is not None
        assert algorithm.name == name
        assert algorithm.display_name == display_name
        assert algorithm.author == author

        manifest = AlgorithmManifest()
        manifest.metadata_required(description="Testing algo manifest builder")
        manifest.interface_required(interface_type=InterfaceType.FILESYSTEM_TASK_WORKER.value)
        manifest.inputs_add_data_type(data_type_name=DataType.PINGS.value,
                                      min_count=2,
                                      max_count=2)

        manifest.container_parameters_required(image="orbitalinsight/raw_foottraffic:84c76f7f",
                                               command=["python",
                                                        "/orbital/base/algorithms/"
                                                        "raw_foottraffic/src/py/raw_foottraffic/simple_foottraffic.py"])
        manifest.outputs_add_data_type(data_type_name="device_visits",
                                       observation_value_columns=["unique_device_count"])

        algorithm_version = await sdk.algorithm_version.create(algorithm_id=algorithm.id,
                                                               algorithm_manifest=manifest)

        assert algorithm_version.id is not None
        await sdk.algorithm_version.deprecate([algorithm_version.id])
        algorithm_versions = await sdk.algorithm_version.get([algorithm_version.id])
        assert algorithm_versions[0].is_deprecated

    @pytest.mark.asyncio
    async def test_deactivate(self):
        sdk = ElementsSDK()
        name = "device-visits-sdk-integration-test-{}".format(uuid.uuid4())
        display_name = "Device Visits SDK Integration"
        author = "elements.sdk.integration@orbitalinsight.com"

        algorithm = await sdk.algorithm.create(name=name,
                                               display_name=display_name,
                                               author=author)
        assert algorithm.id is not None
        assert algorithm.name == name
        assert algorithm.display_name == display_name
        assert algorithm.author == author

        manifest = AlgorithmManifest()
        manifest.metadata_required(description="Testing algo manifest builder")
        manifest.interface_required(interface_type=InterfaceType.FILESYSTEM_TASK_WORKER.value)
        manifest.inputs_add_data_type(data_type_name=DataType.PINGS.value,
                                      min_count=2,
                                      max_count=2)

        manifest.container_parameters_required(image="orbitalinsight/raw_foottraffic:84c76f7f",
                                               command=["python",
                                                        "/orbital/base/algorithms/"
                                                        "raw_foottraffic/src/py/raw_foottraffic/simple_foottraffic.py"])
        manifest.outputs_add_data_type(data_type_name="device_visits",
                                       observation_value_columns=["unique_device_count"])

        algorithm_version = await sdk.algorithm_version.create(algorithm_id=algorithm.id,
                                                               algorithm_manifest=manifest)

        assert algorithm_version.id is not None
        await sdk.algorithm_version.deactivate([algorithm_version.id])
        algorithm_versions = await sdk.algorithm_version.get([algorithm_version.id])
        assert algorithm_versions[0].is_deactivated


class TestAlgorithmConfig:

    @pytest.mark.asyncio
    async def test_create(self):
        sdk = ElementsSDK()
        name = "device-visits-sdk-integration-test-{}".format(uuid.uuid4())
        display_name = "Device Visits SDK Integration"
        author = "elements.sdk.integration@orbitalinsight.com"

        algorithm = await sdk.algorithm.create(name=name,
                                               display_name=display_name,
                                               author=author)
        assert algorithm.id is not None
        assert algorithm.name == name
        assert algorithm.display_name == display_name
        assert algorithm.author == author

        manifest = AlgorithmManifest()
        manifest.metadata_required(description="Testing algo manifest builder")
        manifest.interface_required(interface_type=InterfaceType.FILESYSTEM_TASK_WORKER.value)
        manifest.inputs_add_data_type(data_type_name=DataType.PINGS.value,
                                      min_count=1,
                                      max_count=1)
        manifest.grouping_frequency("DAILY", 1)
        manifest.container_parameters_required(image="orbitalinsight/device_visits:5a579e59",
                                               command=["python",
                                                        "/orbital/base/algorithms/device_visits/src/py/"
                                                        "device_visits/device_visits.py"])
        manifest.outputs_add_data_type(data_type_name="device_visits",
                                       observation_value_columns=["unique_device_count"])

        manifest.parameter_add(name="look_back_time", type="integer", default=3600,
                               description="how far to look back")
        manifest.parameter_add(name="look_forward_time", type="integer", default=3600,
                               description="how far to look back")

        manifest.container_parameters_resource_request(memory_gb=5, cpu_millicore=200)
        algorithm_version = await sdk.algorithm_version.create(algorithm_id=algorithm.id,
                                                               algorithm_manifest=manifest)

        assert algorithm_version.id is not None

        configuration = AlgorithmConfiguration()
        configuration.add_data_source(DataType.PINGS.value, DataSource.WEJO)
        configuration.grouping_frequency(Frequency.WEEKLY, 1)
        algorithm_configuration = await sdk.algorithm_config.create(algorithm_version_id=algorithm_version.id,
                                                                    name="device_visit_sdk_test_config",
                                                                    description="A test configuration.",
                                                                    algorithm_config=configuration)

        assert algorithm_configuration.id is not None

    @pytest.mark.asyncio
    async def test_update(self):
        sdk = ElementsSDK()
        name = "device-visits-sdk-integration-test-{}".format(uuid.uuid4())
        display_name = "Device Visits SDK Integration"
        author = "elements.sdk.integration@orbitalinsight.com"

        algorithm = await sdk.algorithm.create(name=name,
                                               display_name=display_name,
                                               author=author)
        assert algorithm.id is not None
        assert algorithm.name == name
        assert algorithm.display_name == display_name
        assert algorithm.author == author

        manifest = AlgorithmManifest()
        manifest.metadata_required(description="Testing algo manifest builder")
        manifest.interface_required(interface_type=InterfaceType.FILESYSTEM_TASK_WORKER.value)
        manifest.inputs_add_data_type(data_type_name=DataType.PINGS.value,
                                      min_count=1,
                                      max_count=1)
        manifest.grouping_frequency("DAILY", 1)
        manifest.container_parameters_required(image="orbitalinsight/device_visits:5a579e59",
                                               command=["python",
                                                        "/orbital/base/algorithms/device_visits/src/py/"
                                                        "device_visits/device_visits.py"])
        manifest.outputs_add_data_type(data_type_name="device_visits",
                                       observation_value_columns=["unique_device_count"])

        manifest.parameter_add(name="look_back_time", type="integer", default=3600,
                               description="how far to look back")
        manifest.parameter_add(name="look_forward_time", type="integer", default=3600,
                               description="how far to look back")

        manifest.container_parameters_resource_request(memory_gb=5, cpu_millicore=200)
        algorithm_version = await sdk.algorithm_version.create(algorithm_id=algorithm.id,
                                                               algorithm_manifest=manifest)

        assert algorithm_version.id is not None

        configuration = AlgorithmConfiguration()
        configuration.add_data_source(DataType.PINGS.value, DataSource.WEJO)
        configuration.grouping_frequency(Frequency.WEEKLY, 1)
        algorithm_configuration = await sdk.algorithm_config.create(algorithm_version_id=algorithm_version.id,
                                                                    name="device_visit_sdk_test_config",
                                                                    description="A test configuration.",
                                                                    algorithm_config=configuration)
        assert algorithm_configuration.id is not None

        configuration.grouping_frequency(Frequency.DAILY, 2)

        await sdk.algorithm_config.update(algorithm_config_id=algorithm_configuration.id,
                                          algorithm_config=configuration)

    @pytest.mark.asyncio
    @pytest.mark.parametrize("config_count", [
        1,
        10
    ])
    async def test_get(self, config_count):
        sdk = ElementsSDK()
        name = "device-visits-sdk-integration-test-{}".format(uuid.uuid4())
        display_name = "Device Visits SDK Integration"
        author = "elements.sdk.integration@orbitalinsight.com"

        algorithm = await sdk.algorithm.create(name=name,
                                               display_name=display_name,
                                               author=author)
        assert algorithm.id is not None
        assert algorithm.name == name
        assert algorithm.display_name == display_name
        assert algorithm.author == author

        manifest = AlgorithmManifest()
        manifest.metadata_required(description="Testing algo manifest builder")
        manifest.interface_required(interface_type=InterfaceType.FILESYSTEM_TASK_WORKER.value)
        manifest.inputs_add_data_type(data_type_name=DataType.PINGS.value,
                                      min_count=1,
                                      max_count=1)

        manifest.container_parameters_required(image="orbitalinsight/device_visits:5a579e59",
                                               command=["python",
                                                        "/orbital/base/algorithms/device_visits/src/py/"
                                                        "device_visits/device_visits.py"])
        manifest.outputs_add_data_type(data_type_name="device_visits",
                                       observation_value_columns=["unique_device_count"])

        manifest.parameter_add(name="look_back_time", type="integer", default=3600,
                               description="how far to look back")
        manifest.parameter_add(name="look_forward_time", type="integer", default=3600,
                               description="how far to look back")

        manifest.container_parameters_resource_request(memory_gb=5, cpu_millicore=200)
        algorithm_version = await sdk.algorithm_version.create(algorithm_id=algorithm.id,
                                                               algorithm_manifest=manifest)

        assert algorithm_version.id is not None
        config_ids = []
        for i in range(config_count):
            configuration = AlgorithmConfiguration()
            configuration.add_data_source(DataType.PINGS.value, DataSource.WEJO)

            algorithm_configuration = await sdk.algorithm_config.create(algorithm_version_id=algorithm_version.id,
                                                                        name="device_visit_sdk_test_config",
                                                                        description="A test configuration.",
                                                                        algorithm_config=configuration)

            assert algorithm_configuration.id is not None
            config_ids.append(algorithm_configuration.id)

        configurations = await sdk.algorithm_config.get(config_ids, pagination=Pagination(page_size=5))
        assert len(config_ids) == config_count == len(configurations)

    @pytest.mark.asyncio
    @pytest.mark.parametrize("config_count", [
        10,
        20
    ])
    async def test_list(self, config_count: int):
        sdk = ElementsSDK()
        name = "device-visits-sdk-integration-test-{}".format(uuid.uuid4())
        display_name = "Device Visits SDK Integration"
        author = "elements.sdk.integration@orbitalinsight.com"

        algorithm = await sdk.algorithm.create(name=name,
                                               display_name=display_name,
                                               author=author)
        assert algorithm.id is not None
        assert algorithm.name == name
        assert algorithm.display_name == display_name
        assert algorithm.author == author

        manifest = AlgorithmManifest()
        manifest.metadata_required(description="Testing algo manifest builder")
        manifest.interface_required(interface_type=InterfaceType.FILESYSTEM_TASK_WORKER.value)
        manifest.inputs_add_data_type(data_type_name=DataType.PINGS.value,
                                      min_count=1,
                                      max_count=1)

        manifest.container_parameters_required(image="orbitalinsight/device_visits:5a579e59",
                                               command=["python",
                                                        "/orbital/base/algorithms/device_visits/src/py/"
                                                        "device_visits/device_visits.py"])
        manifest.outputs_add_data_type(data_type_name="device_visits",
                                       observation_value_columns=["unique_device_count"])

        manifest.parameter_add(name="look_back_time", type="integer", default=3600,
                               description="how far to look back")
        manifest.parameter_add(name="look_forward_time", type="integer", default=3600,
                               description="how far to look back")

        manifest.container_parameters_resource_request(memory_gb=5, cpu_millicore=200)
        algorithm_version = await sdk.algorithm_version.create(algorithm_id=algorithm.id,
                                                               algorithm_manifest=manifest)

        assert algorithm_version.id is not None
        config_ids = []
        search_me = str(uuid.uuid4())
        for i in range(config_count):
            configuration = AlgorithmConfiguration()
            configuration.add_data_source(DataType.PINGS.value, DataSource.WEJO)

            algorithm_configuration = await sdk.algorithm_config.create(algorithm_version_id=algorithm_version.id,
                                                                        name="device_visit_sdk_test_config",
                                                                        description=search_me,
                                                                        algorithm_config=configuration)

            assert algorithm_configuration.id is not None
            config_ids.append(algorithm_configuration.id)

        configurations = await sdk.algorithm_config.list()
        assert len(configurations) > 0
        configurations = await sdk.algorithm_config.list(algorithm_version_id=algorithm_version.id,
                                                         include_all_versions=True)
        assert len(config_ids) == config_count == len(configurations)

    @pytest.mark.asyncio
    @pytest.mark.parametrize("config_count", [
        10
    ])
    async def test_deprecate(self, config_count):
        sdk = ElementsSDK()
        name = "device-visits-sdk-integration-test-{}".format(uuid.uuid4())
        display_name = "Device Visits SDK Integration"
        author = "elements.sdk.integration@orbitalinsight.com"

        algorithm = await sdk.algorithm.create(name=name,
                                               display_name=display_name,
                                               author=author)
        assert algorithm.id is not None
        assert algorithm.name == name
        assert algorithm.display_name == display_name
        assert algorithm.author == author

        manifest = AlgorithmManifest()
        manifest.metadata_required(description="Testing algo manifest builder")
        manifest.interface_required(interface_type=InterfaceType.FILESYSTEM_TASK_WORKER.value)
        manifest.inputs_add_data_type(data_type_name=DataType.PINGS.value,
                                      min_count=1,
                                      max_count=1)

        manifest.container_parameters_required(image="orbitalinsight/device_visits:5a579e59",
                                               command=["python",
                                                        "/orbital/base/algorithms/device_visits/src/py/"
                                                        "device_visits/device_visits.py"])
        manifest.outputs_add_data_type(data_type_name="device_visits",
                                       observation_value_columns=["unique_device_count"])

        manifest.parameter_add(name="look_back_time", type="integer", default=3600,
                               description="how far to look back")
        manifest.parameter_add(name="look_forward_time", type="integer", default=3600,
                               description="how far to look back")

        manifest.container_parameters_resource_request(memory_gb=5, cpu_millicore=200)
        algorithm_version = await sdk.algorithm_version.create(algorithm_id=algorithm.id,
                                                               algorithm_manifest=manifest)

        assert algorithm_version.id is not None
        config_ids = []
        search_me = str(uuid.uuid4())
        for i in range(config_count):
            configuration = AlgorithmConfiguration()
            configuration.add_data_source(DataType.PINGS.value, DataSource.WEJO)

            algorithm_configuration = await sdk.algorithm_config.create(algorithm_version_id=algorithm_version.id,
                                                                        name="device_visit_sdk_test_config",
                                                                        description=search_me,
                                                                        algorithm_config=configuration)

            assert algorithm_configuration.id is not None
            config_ids.append(algorithm_configuration.id)

        await sdk.algorithm_config.deprecate(algorithm_config_ids=config_ids)
        configurations = await sdk.algorithm_config.get(ids=config_ids)
        for config in configurations:
            assert config.is_deprecated

    @pytest.mark.asyncio
    @pytest.mark.parametrize("config_count", [
        10
    ])
    async def test_deactivate(self, config_count):
        sdk = ElementsSDK()
        name = "device-visits-sdk-integration-test-{}".format(uuid.uuid4())
        display_name = "Device Visits SDK Integration"
        author = "elements.sdk.integration@orbitalinsight.com"

        algorithm = await sdk.algorithm.create(name=name,
                                               display_name=display_name,
                                               author=author)
        assert algorithm.id is not None
        assert algorithm.name == name
        assert algorithm.display_name == display_name
        assert algorithm.author == author

        manifest = AlgorithmManifest()
        manifest.metadata_required(description="Testing algo manifest builder")
        manifest.interface_required(interface_type=InterfaceType.FILESYSTEM_TASK_WORKER.value)
        manifest.inputs_add_data_type(data_type_name=DataType.PINGS.value,
                                      min_count=1,
                                      max_count=1)

        manifest.container_parameters_required(image="orbitalinsight/device_visits:5a579e59",
                                               command=["python",
                                                        "/orbital/base/algorithms/device_visits/src/py/"
                                                        "device_visits/device_visits.py"])
        manifest.outputs_add_data_type(data_type_name="device_visits",
                                       observation_value_columns=["unique_device_count"])

        manifest.parameter_add(name="look_back_time", type="integer", default=3600,
                               description="how far to look back")
        manifest.parameter_add(name="look_forward_time", type="integer", default=3600,
                               description="how far to look back")

        manifest.container_parameters_resource_request(memory_gb=5, cpu_millicore=200)
        algorithm_version = await sdk.algorithm_version.create(algorithm_id=algorithm.id,
                                                               algorithm_manifest=manifest)

        assert algorithm_version.id is not None
        config_ids = []
        search_me = str(uuid.uuid4())
        for i in range(config_count):
            configuration = AlgorithmConfiguration()
            configuration.add_data_source(DataType.PINGS.value, DataSource.WEJO)

            algorithm_configuration = await sdk.algorithm_config.create(algorithm_version_id=algorithm_version.id,
                                                                        name="device_visit_sdk_test_config",
                                                                        description=search_me,
                                                                        algorithm_config=configuration)

            assert algorithm_configuration.id is not None
            config_ids.append(algorithm_configuration.id)

        await sdk.algorithm_config.deactivate(algorithm_config_ids=config_ids)
        configurations = await sdk.algorithm_config.get(ids=config_ids)
        for config in configurations:
            assert config.is_deactivated

    @pytest.mark.asyncio
    async def test_delete(self):
        pass


class TestAlgorithmComputation:

    @pytest.mark.asyncio
    async def test_create(self):
        sdk = ElementsSDK()

        date_format = "%Y-%m-%dT%H:%M:%SZ"
        toi_configuration = TOIBuilder()
        toi_configuration.build_toi(start=datetime.strptime("2019-01-05T01:00:00Z", date_format),
                                    finish=datetime.strptime("2019-01-06T01:00:00Z", date_format))
        toi_configuration.build_recurrence(TOIRuleBuilder.build_rule(
            frequency=Frequency.DAILY,
            interval=1))
        toi = await sdk.toi.create(toi_configuration.get())

        aoi_collection = await sdk.aoi_collection.create(aoi_collection_name=str(uuid.uuid4()))
        await sdk.aoi.upload(aoi_collection_id=aoi_collection.id,
                             file_path="../resources/aois/geojson/basic_factory_phillips_66.geojson")

        device_visit_algorithm = await sdk.algorithm.create(name="device-visits-{}".format(uuid.uuid4()),
                                                            author="sdk@orbitalinsight.com",
                                                            display_name="Device Visits")

        logging.info(device_visit_algorithm.id)
        manifest = AlgorithmManifest()
        manifest.metadata_required(description="Testing algo manifest builder")
        manifest.interface_required(interface_type=InterfaceType.FILESYSTEM_TASK_WORKER.value)
        manifest.inputs_add_data_type(data_type_name=DataType.PINGS.value)
        manifest.container_parameters_required(image="orbitalinsight/device_visits:5a579e59",
                                               command=["python",
                                                        "/orbital/base/algorithms/device_visits/src/py/"
                                                        "device_visits/device_visits.py"])
        manifest.outputs_add_data_type(data_type_name="device_visits",
                                       observation_value_columns=["unique_device_count"])
        manifest.parameter_add(name="look_back_time", type="integer", default=3600,
                               description="how far to look back")
        manifest.parameter_add(name="look_forward_time", type="integer", default=3600,
                               description="how far to look back")
        manifest.container_parameters_resource_request(memory_gb=5, cpu_millicore=200)
        device_visit_algorithm_version = await sdk.algorithm_version.create(algorithm_id=device_visit_algorithm.id,
                                                                            algorithm_manifest=manifest)

        config = AlgorithmConfiguration()
        config.add_data_source(data_type=DataType.PINGS.value, data_source=DataSource.WEJO)

        device_visit_algorithm_config = await sdk.algorithm_config.create(
            algorithm_version_id=device_visit_algorithm_version.id,
            name="device_visit_test_{}".format(uuid.uuid4()), description="a description", algorithm_config=config)

        algorithm_computation = await sdk.algorithm_computation.create(
            aoi_collection_id=aoi_collection.id,
            toi_id=toi.id,
            algorithm_config_id=device_visit_algorithm_config.id)

        assert algorithm_computation.id is not None

        algorithm_computation = await sdk.algorithm_computation.get([algorithm_computation.id])
        print(algorithm_computation)
        # status = ""
        # while status != "COMPLETED":
        #     pass

    @pytest.mark.asyncio
    async def test_run(self):
        sdk = ElementsSDK()

        datetime_format = "%Y-%m-%dT%H:%M:%SZ"
        toi_configuration = TOIBuilder()
        toi_configuration.build_toi(start=datetime.strptime("2019-01-05T01:00:00Z", datetime_format),
                                    finish=datetime.strptime("2019-01-06T01:00:00Z", datetime_format))
        toi_configuration.build_recurrence(TOIRuleBuilder.build_rule(
            frequency=Frequency.DAILY,
            interval=1
        ))
        toi = await sdk.toi.create(toi_configuration.get())

        aoi_collection = await sdk.aoi_collection.create(aoi_collection_name=str(uuid.uuid4()))
        await sdk.aoi.upload(aoi_collection_id=aoi_collection.id,
                             file_path="../resources/aois/geojson/basic_factory_phillips_66.geojson")

        device_visit_algorithm = await sdk.algorithm.create(name="device-visits-{}".format(uuid.uuid4()),
                                                            author="sdk@orbitalinsight.com",
                                                            display_name="Device Visits")

        manifest = AlgorithmManifest()
        manifest.metadata_required(description="Testing algo manifest factory")
        manifest.interface_required(interface_type=InterfaceType.FILESYSTEM_TASK_WORKER.value)
        manifest.inputs_add_data_type(data_type_name=DataType.PINGS.value)
        manifest.container_parameters_required(image="orbitalinsight/device_visits:5a579e59",
                                               command=["python",
                                                        "/orbital/base/algorithms/device_visits/src/py/"
                                                        "device_visits/device_visits.py"])
        manifest.outputs_add_data_type(data_type_name="device_visits",
                                       observation_value_columns=["unique_device_count"])
        manifest.parameter_add(name="look_back_time", type="integer", default=3600,
                               description="how far to look back")
        manifest.parameter_add(name="look_forward_time", type="integer", default=3600,
                               description="how far to look back")
        manifest.container_parameters_resource_request(memory_gb=5, cpu_millicore=200)
        device_visit_algorithm_version = await sdk.algorithm_version.create(algorithm_id=device_visit_algorithm.id,
                                                                            algorithm_manifest=manifest)

        config = AlgorithmConfiguration()
        config.add_data_source(data_type=DataType.PINGS.value, data_source=DataSource.WEJO)

        device_visit_algorithm_config = await sdk.algorithm_config.create(
            algorithm_version_id=device_visit_algorithm_version.id,
            name="device_visit_test_{}".format(uuid.uuid4()), description="a description", algorithm_config=config)

        algorithm_computation = await sdk.algorithm_computation.create(
            aoi_collection_id=aoi_collection.id,
            toi_id=toi.id,
            algorithm_config_id=device_visit_algorithm_config.id)

        assert algorithm_computation.id is not None

        # RUN ALGORITHM COMPUTATION
        await sdk.algorithm_computation.run([algorithm_computation.id])

    @pytest.mark.asyncio
    async def test_get(self):
        sdk = ElementsSDK()

        datetime_format = '%Y-%m-%dT%H:%M:%SZ'
        toi_configuration = TOIBuilder()
        toi_configuration.build_toi(start=datetime.strptime("2019-01-05T01:00:00Z", datetime_format),
                                    finish=datetime.strptime("2019-01-06T01:00:00Z", datetime_format))
        toi_configuration.build_recurrence(TOIRuleBuilder.build_rule(
            frequency=Frequency.DAILY,
            interval=1
        ))
        toi = await sdk.toi.create(toi_configuration.get())

        aoi_collection = await sdk.aoi_collection.create(aoi_collection_name=str(uuid.uuid4()))
        await sdk.aoi.upload(aoi_collection_id=aoi_collection.id,
                             file_path="../resources/aois/geojson/basic_factory_phillips_66.geojson")

        device_visit_algorithm = await sdk.algorithm.create(name="device-visits-{}".format(uuid.uuid4()),
                                                            author="sdk@orbitalinsight.com",
                                                            display_name="Device Visits")

        manifest = AlgorithmManifest()
        manifest.metadata_required(description="Testing algo manifest factory")
        manifest.interface_required(interface_type=InterfaceType.FILESYSTEM_TASK_WORKER.value)
        manifest.inputs_add_data_type(data_type_name=DataType.PINGS.value)
        manifest.container_parameters_required(image="orbitalinsight/device_visits:5a579e59",
                                               command=["python",
                                                        "/orbital/base/algorithms/device_visits/src/py/"
                                                        "device_visits/device_visits.py"])
        manifest.outputs_add_data_type(data_type_name="device_visits",
                                       observation_value_columns=["unique_device_count"])
        manifest.parameter_add(name="look_back_time", type="integer", default=3600,
                               description="how far to look back")
        manifest.parameter_add(name="look_forward_time", type="integer", default=3600,
                               description="how far to look back")
        manifest.container_parameters_resource_request(memory_gb=5, cpu_millicore=200)
        device_visit_algorithm_version = await sdk.algorithm_version.create(algorithm_id=device_visit_algorithm.id,
                                                                            algorithm_manifest=manifest)

        config = AlgorithmConfiguration()
        config.add_data_source(data_type=DataType.PINGS.value, data_source=DataSource.WEJO)

        device_visit_algorithm_config = await sdk.algorithm_config.create(
            algorithm_version_id=device_visit_algorithm_version.id,
            name="device_visit_test_{}".format(uuid.uuid4()), description="a description", algorithm_config=config)

        algorithm_computation = await sdk.algorithm_computation.create(
            aoi_collection_id=aoi_collection.id,
            toi_id=toi.id,
            algorithm_config_id=device_visit_algorithm_config.id)

        assert algorithm_computation.id is not None

        # RUN ALGORITHM COMPUTATION
        await sdk.algorithm_computation.run([algorithm_computation.id])

        algorithm_computation = await sdk.algorithm_computation.get([algorithm_computation.id])
        assert algorithm_computation is not None



from typing import List, Dict
from google.protobuf.struct_pb2 import Struct

from elements_api import ElementsAsyncClient
from elements_api.models.data_source_pb2 import DataSourceGetRequest, DataSourceListRequest, DataSource, DataSourceCreateRequest
from elements_api.models.data_type_pb2 import DataTypeGetRequest, DataTypeListRequest, DataType, DataTypeCreateRequest


class APIDataSource:
    def __init__(self, client: ElementsAsyncClient, timeout):
        self.__timeout = timeout
        self.__client = client

    async def get(self, ids: List[str]) -> List[DataSource]:
        """
        Retrieves all the details about the specified data source.
        :param ids:
        :return: List[DataSource]
        """
        request = DataSourceGetRequest(
            ids=ids
        )
        response = await self.__client.api.data_source.get(request, timeout=self.__timeout)
        return response.data_sources

    async def list(self, **kwargs) -> List[DataSource]:
        """
        List all available data sources that the user has access to.
        :param kwargs:
            - search_text: str
        :return: List[DataSource]
        """
        message_fragments = []
        for kwarg in kwargs.keys():
            if 'search_text' == kwarg:
                message_fragments.append(DataSourceListRequest(
                    search_text=kwargs[kwarg]
                ))
        request = DataSourceListRequest()
        for fragment in message_fragments:
            request.MergeFrom(fragment)

        response = await self.__client.api.data_source.list(request, timeout=self.__timeout)
        return response.data_sources

    async def create(self, data_source: Dict):
        data_type_objs = [DataType(name=name) for name in data_source.get("data_types", [])]
        config_info_struct = Struct()
        if "config_info" in data_source and isinstance(data_source["config_info"], dict):
            config_info_struct.update(data_source["config_info"])
        data_source_create_request = DataSourceCreateRequest(
            id=data_source.get("id", ""),
            name=data_source.get("name", ""),
            description=data_source.get("description", ""),
            data_types=data_type_objs,
            config_info=config_info_struct
        )
        response = await self.__client.api.data_source.create(data_source_create_request, timeout=self.__timeout)
        return response


class APIDataType:
    def __init__(self, client: ElementsAsyncClient, timeout):
        self.__timeout = timeout
        self.__client = client

    async def create(self, name: str, description: str, schema: str, data_source_ids: List,
                     sensor_type: str) -> DataType:
        """

        :param name:
        :param description:
        :param schema:
        :param sensor_type:
        :return:
        """
        request = DataTypeCreateRequest(
            name=name,
            description=description,
            schema=schema,
            sensor_type=sensor_type
        )
        response = await self.__client.api.data_type.create(request)
        return response.data_type

    async def get(self, names: List[str]) -> List[DataType]:
        """
        Gets a data type that the user has access to.

        :param ids: These are the name fields for the data source
        :return: List[DataType]
        """
        request = DataTypeGetRequest(
            names=names
        )
        response = await self.__client.api.data_type.get(request, timeout=self.__timeout)
        return response.data_types

    async def list(self, **kwargs) -> List[DataType]:
        """
        List all registered data_types.

        :param kwargs:
            - search_text: str
        :return: List[DataType]
        """
        message_fragments = []
        for kwarg in kwargs.keys():
            if 'search_text' == kwarg:
                message_fragments.append(DataTypeListRequest(
                    search_text=kwargs[kwarg]
                ))
        request = DataTypeListRequest()
        for fragment in message_fragments:
            request.MergeFrom(fragment)

        response = await self.__client.api.data_type.list(request, timeout=self.__timeout)
        return response.data_types



import click
import asyncio
import pandas as pd
import yaml
import os
import json
import elements.cli.lib.utils as tsu
from elements.cli.lib.aliased_group import AliasedGroup
import elements.cli.lib.workflow as wf
from google.protobuf.json_format import MessageToDict


@click.command(cls=AliasedGroup, help="Algorithm 'version' command group")
@click.pass_context
def version(ctx):
    pass


@version.command('get')
@click.pass_context
@click.option('-ia', '--algorithm_id', type=str,
              help="UUID of the algorithm to look up the algorithm_versions.")
@click.option('-iv', '--algorithm_version_id', type=str,
              help="UUID of the algorithm_version to look up.")
@click.option('-n', '--algorithm_name', type=str,
              help="Name of the algorithm to look up the algorithm_versions.  All matching algos will be queries")
@click.option('-C', '--only_columns', multiple=True,
              help="Include only the listed columns in output.")
@click.option('-c', '--append_columns', multiple=True,
              help="Append these columns to those shown by default.")
@click.option('-T', '--transpose', is_flag=True,
              help="Display the result in transpose (helpful to show algorithms with long string values).")
@click.option('-om', '--output_manifest', type=str,
              help="Write manifest to this filename.", default=None)
def algorithm_version_get(
        ctx,
        algorithm_id,
        algorithm_version_id,
        algorithm_name,
        only_columns,
        append_columns,
        transpose,
        output_manifest=None,
):
    columns = ['id', 'created_on', 'algorithm.name', 'algorithm.author']
    if only_columns:
        columns = list(only_columns)
    if append_columns:
        columns = columns + list(append_columns)

    # if they provide an explicit algo version ID, get it.
    algo_versions = []
    if algorithm_version_id:
        algo_versions += asyncio.run(wf.get_algorithm_versions(algorithm_version_id=algorithm_version_id))

    # if they provide a algorithm_id or name, look up the algo versions and get them.
    algorithm_ids = []
    if algorithm_id is not None:
        algorithm_ids = [algorithm_id]
    if algorithm_name is not None:
        algos = asyncio.run(wf.list_algorithms(algorithm_name))
        algorithm_ids += [algo.id for algo in algos]

    for algorithm_id in algorithm_ids:
        algo_versions += asyncio.run(wf.get_algorithm_versions(algorithm_id=algorithm_id))

    # print the results
    if algo_versions:
        data = []
        for algo_version in algo_versions:
            algo_version_dict = tsu.protobuf_to_dict(algo_version)
            data.append(algo_version_dict)

        df = pd.DataFrame(data=data)
        if 'all' not in columns:
            df = df[tsu.match_columns(df.columns, columns)]
        if transpose:
            df = df.T
        df = df.rename(columns={'id': 'algorithm_version_id'})
        tsu.set_pandas_display()
        click.secho(f"algorithm_id: {algorithm_id}", fg='cyan')
        click.echo(df)

    if output_manifest:
        n_versions = len(algo_versions)
        for algo_version in algo_versions:

            manifest = MessageToDict(algo_version.manifest)
            algorithm_id = algo_version.algorithm.id
            manifest['algorithm_id'] = algorithm_id
            manifest['algorithm_name'] = algo_version.algorithm.name

            output_manifest_fname = output_manifest
            if n_versions > 1:
                fname, ext = os.path.splitext(output_manifest)
                output_manifest_fname = f"{fname}_{algorithm_id[:8]}.{ext}"
            with open(output_manifest_fname, 'w') as fp:
                if output_manifest.endswith('json'):
                    json.dump(manifest, fp, indent=4)
                else:
                    yaml.dump(manifest, fp)


@version.command('create')
@click.pass_context
@click.option('-ia', '--algorithm_id', type=str)
@click.option('-m', '--manifest_yaml', type=str)
@click.option('-d', '--docker_version_hash', type=str)
@click.option('--dry_run', is_flag=True, default=False)
def algorithm_version_create(
        ctx,
        algorithm_id,
        manifest_yaml,
        docker_version_hash,
        dry_run,
):

    wf.check_environment_complete(raise_on_failure=True)

    # create the manifest
    input_manifest = None
    with open(manifest_yaml, 'r') as fp:
        input_manifest = yaml.safe_load(fp)

    image = wf.update_manifest_docker_hash(input_manifest, docker_version_hash)
    manifest = wf.create_manifest(input_manifest)

    click.echo("Using the following information:")
    click.echo(f"--> algorithm_id:  {algorithm_id}")
    click.echo(f"--> image:         {image}")

    # create the manifest
    if not dry_run:
        algorithm_version = asyncio.run(wf.create_algorithm_version(
            algorithm_id,
            manifest,
        ))

        click.echo(f"Your new algorithm_version_id is: {algorithm_version.id}")


@version.command('deactivate')
@click.pass_context
@click.option('-iv', '--algorithm_version_ids', type=str, multiple=True, required=True,
              help="UUIDs of the algorithm_versions to deactivate")
def algorithm_version_deactivate(
        ctx,
        algorithm_version_ids,
):

    response = asyncio.run(wf.deactivate_algorithm_versions(algorithm_version_ids=algorithm_version_ids))

    click.echo("Deactivation response code: ", nl=False)
    if response:
        click.secho(f" {response} (FAILURE)", fg='red')
    else:
        click.secho(f" {response} (SUCCESS)", fg='green')



import logging
import uuid
from typing import List

import pytest

from elements.sdk.api.aoi import AOIInputBuilder
from elements.sdk.elements_sdk import ElementsSDK


class TestAOI:

    @pytest.mark.asyncio
    async def test_upload(self):

        sdk = ElementsSDK()
        aoi_collection = await sdk.aoi_collection.create(aoi_collection_name=str(uuid.uuid4()))

        assert aoi_collection.id is not None

        transaction = await sdk.aoi.upload(aoi_collection_id=aoi_collection.id,
                                           file_path="../resources/aois/geojson/us-amz-distro-centers.geojson")

        logging.info(transaction)

    @pytest.mark.asyncio
    @pytest.mark.parametrize("name, category, category_type, tags", [
        ("test_name-{}".format(uuid.uuid4()), "industrial", "manufacturing", ["aoi_tag1", "aoi_tag2"]),
        ("1-test_name-{}".format(uuid.uuid4()), "residential", "apartments", ["aoi_tag1", "aoi_tag2"])])
    async def test_create(self,
                          name: str,
                          category: str,
                          category_type: str,
                          tags: List):
        sdk = ElementsSDK()

        aoi_collection = await sdk.aoi_collection.create(aoi_collection_name=name)

        assert aoi_collection.id is not None

        wkt = """POLYGON ((-121.891336 37.345116, -121.882978 37.322622, -121.865618 37.335404, -121.891336 37.345116))
        """
        aoi_builder = AOIInputBuilder()
        aoi = aoi_builder.build(geom_wkt=wkt,
                                name="aoi-" + name,
                                category=category,
                                type=category_type,
                                tags=['sdk-integration'],
                                # attributes={"organization": "google", "city": "Palo Alto"}
                                )
        aoi_identifiers = await sdk.aoi.create(aoi_collection_id=aoi_collection.id,
                                               aoi_inputs=[aoi])

        assert aoi_identifiers is not None

    @pytest.mark.asyncio
    @pytest.mark.parametrize("name, category, category_type, tags", [
        ("test_name-{}".format(uuid.uuid4()), "industrial", "manufacturing", ["aoi_tag1", "aoi_tag2"]),
        ("1-test_name-{}".format(uuid.uuid4()), "residential", "apartments", ["aoi_tag1", "aoi_tag2"])])
    async def test_get(self,
                       name: str,
                       category: str,
                       category_type: str,
                       tags: List):
        sdk = ElementsSDK()

        aoi_collection = await sdk.aoi_collection.create(aoi_collection_name=name)

        assert aoi_collection.id is not None

        wkt = """POLYGON ((-121.891336 37.345116, -121.882978 37.322622, -121.865618 37.335404, -121.891336 37.345116))
                """

        aoi_builder = AOIInputBuilder()
        aoi = aoi_builder.build(geom_wkt=wkt,
                                name="aoi-" + name,
                                category=category,
                                type=category_type,
                                tags=['sdk-integration'])
        aoi_identifiers = await sdk.aoi.create(aoi_collection_id=aoi_collection.id,
                                               aoi_inputs=[aoi])

        response = sdk.aoi.get(ids=aoi_identifiers, verbose=True)

        assert response is not None


class TestAOIVersion:
    @pytest.mark.asyncio
    @pytest.mark.parametrize("name, category, category_type, tags", [
        ("test_name-{}".format(uuid.uuid4()), "industrial", "manufacturing", ["aoi_tag1", "aoi_tag2"]),
        ("1-test_name-{}".format(uuid.uuid4()), "residential", "apartments", ["aoi_tag1", "aoi_tag2"])])
    async def test_create(self,
                          name: str,
                          category: str,
                          category_type: str,
                          tags: List):
        sdk = ElementsSDK()

        aoi_collection = await sdk.aoi_collection.create(aoi_collection_name=name)

        assert aoi_collection.id is not None

        wkt = """POLYGON ((-121.891336 37.345116, -121.882978 37.322622, -121.865618 37.335404, -121.891336 37.345116))
        """
        aoi_builder = AOIInputBuilder()
        aoi = aoi_builder.build(geom_wkt=wkt,
                                name="aoi-" + name,
                                category=category,
                                type=category_type,
                                tags=['sdk-integration'],
                                # attributes={"organization": "google", "city": "Palo Alto"}
                                )
        aoi_identifiers = await sdk.aoi.create(aoi_collection_id=aoi_collection.id,
                                               aoi_inputs=[aoi])

        assert aoi_identifiers is not None

        aoi_version_id = sdk.aoi_version.create(aoi_identifiers[0], aoi_modification_input=aoi)

        assert aoi_version_id is not None

    @pytest.mark.asyncio
    async def test_get(self):
        print("test needs to be written")
        pass

    @pytest.mark.asyncio
    async def test_list(self):
        print("test needs to be written")
        pass


class TestAOICollection:
    @pytest.mark.asyncio
    @pytest.mark.parametrize("name", [("test_name-{}".format(uuid.uuid4())),
                                      ("1-test_name-{}".format(uuid.uuid4()))])
    async def test_create(self, name):
        sdk = ElementsSDK()

        aoi_collection = await sdk.aoi_collection.create(aoi_collection_name=name)

        assert aoi_collection.id is not None


    @pytest.mark.asyncio
    async def test_get(self):
        sdk = ElementsSDK()

        aoi_collection = await sdk.aoi_collection.create(aoi_collection_name=str(uuid.uuid4()))

        assert aoi_collection.id is not None

        await sdk.aoi.upload(aoi_collection_id=aoi_collection.id,
                             file_path="../resources/aois/geojson/us-amz-distro-centers.geojson")

        aoi_collection = await sdk.aoi_collection.get(aoi_collection_id=aoi_collection.id)
        print(aoi_collection)





import logging
import uuid
from typing import List

from elements_api.models.analysis_pb2 import AnalysisAlgorithmNode, AnalysisAlgorithmConfigNode
from elements_api.models.analysis_version_pb2 import AnalysisManifest as AnalysisManifestPAPI

fmt = "%(levelname)s %(asctime)s %(filename)s:%(lineno)i %(funcName)s pid:%(process)s: %(message)s"
logging.basicConfig(
    level=logging.INFO,
    format=fmt,
    datefmt="%H:%M:%S",
)


class AnalysisManifest:
    """
    Description:
    This will be a builder to build analyses in an easy and controlled way. Specifically this will manage the sequence
    of algorithms needed for a analyses. The object structure will be maintained in way that once the build is complete
    the returned object can be submitted to the analyses end point.

    Example:
        Note - dag1 can be any name you want, but each time you call analysesDAGFactory() that creates a new "instance"
            that instance MUST ONLY be used for 1 DAG. If you want to make another tag call the above on a new variable.

        dag1 = analysesDAGFactory()

        dag1.add_node("algo1", "7bd14fed-a6d7-42f7-b1e3-a9b1043ebb16")
        dag1.add_node("algo2", "aa643c56-9420-4707-a7da-bf01f89e18bd")
        dag1.add_node("algo3", "707d9325-a4e6-4d32-8991-47962e1337e8")

    """

    def __init__(self):
        self.__manifest = AnalysisManifestPAPI()

        # Set Default Manifest Version
        self.__manifest.manifest_version = "0.1.0"

    def metadata(self, description: str, tags: List):
        self.__manifest.metadata.description = description
        self.__manifest.metadata.tags.extend(tags)

    def manifest_version(self, version: str):
        self.__manifest.manifest_version = version

    def add_node(self, name: str, algorithm_version_id: str, children: [] = None):
        """
        Description:
            Add a node to the dag. Note, this is a unique name safe generated as it will add a UUID to the end
            of the name given as a suffix to ensure uniqueness.

        :param name: A unique (relative to this analyses) name for this node
        :param algorithm_version_id: The version id of the algorithm that will be executed at this node.
        :param children: these are children algorithms to run from this node.

                     Node1
               //      ||      \\
            Child1   Child2    ...
               \\      ||      //
                     Node2
                       ||
                       ...
        :return:
        """
        if not children:
            children = []
        for node in self.__manifest.algorithm_nodes:
            assert name != node.name
        self.__manifest.algorithm_nodes.append(AnalysisAlgorithmNode(
            name=name,
            algorithm_version_id=algorithm_version_id,
            children=children
        ))

    def add_node_edge(self, parent_index: int, child_index: int):
        """
        Creates a directed edge between two nodes in the dag
        :param parent_index: The parent node index to draw an edge from
        :param child_index: The child node index to draw an edge to
        :return:
        """
        assert 0 <= parent_index <= len(self.__manifest.algorithm_nodes) - 1
        self.__manifest.algorithm_nodes[parent_index].children.append(
            self.__manifest.algorithm_nodes[child_index].name)

    def add_child_node(self, node_index: int, name: str, algorithm_config_id: str):
        """
        Description
         This will be future implementation. Right now child nodes are edges, use alternative method.
        TODO: This is still WIP. Nailing down specifics on child nodes and conditions for add or not.
        :param node_index:
        :param name:
        :param algorithm_config_id:
        :return:
        """
        assert 0 <= node_index <= len(self.__manifest.algorithm_nodes) - 1
        self.__manifest.algorithm_nodes[node_index]['children'].append((
            AnalysisAlgorithmNode(
                name="{}-{}".format(name, uuid.uuid4()),
                algorithm_version_id=algorithm_config_id
            )
        ))

    def get_name(self, node_index: int):
        """
        Description:

        Gets the node name at a specific index.

        :param node_index:
        :return:
        """
        if 0 <= node_index <= len(self.__manifest.algorithm_nodes):
            return self.__manifest.algorithm_nodes[node_index].name

    def get(self) -> AnalysisManifestPAPI:
        return self.__manifest

    # Helper Methods
    def print(self):
        """
        Description
        Helper method to see your graph structure. Not to be used in production code.

        :return:
        """
        for i in range(len(self.__manifest.algorithm_nodes)):
            logging.info("Node-{}:\n\tname: {}\n\talgorithm_version_id: {}\n\tchildren:"
                         .format(i,
                                 self.__manifest.algorithm_nodes[i].name,
                                 self.__manifest.algorithm_nodes[i].algorithm_version_id))
            if self.__manifest.algorithm_nodes[i].children:
                for edge in self.__manifest.algorithm_nodes[i].children:
                    logging.info("\t\tchild_name: {}".format(edge))
            else:
                logging.info("\t\tEmpty")


class AnalysisConfiguration:
    """
    Description:
    This builder will construct the Analysis dag. Use similar to AnalysisDAGFactory.

    """

    def __init__(self, analysis_version_id: str):
        self.__analysis_version_id = analysis_version_id
        self.__analysis_config_nodes = []

    def add_config_node(self, name: str, algorithm_config_id: str, index: int = None):
        """
        Description:
        This method adds an AnalysisAlgorithmConfigNode which references the EXACT node name used to define a node in
        the analyses that you are referencing. It also assigns the Algorithm Configuration ID to that node meaning
        here are the specific runtime specifications.

        :param index: optional - if specified, the node will be inserted before the provided index.
        :param name: The name of the node in the analyses you wish to set the config for. Exact Match required.
        :param algorithm_config_id: The algorithm configuration ID to assign to that node.
        :return:
        """
        if index is None:
            self.__analysis_config_nodes.append(AnalysisAlgorithmConfigNode(
                name=name,
                algorithm_config_id=algorithm_config_id
            ))
        else:
            if not 0 <= index < len(self.__analysis_config_nodes):
                logging.info("Provided index {} is output of bounds".format(index))
            self.__analysis_config_nodes.insert(index, AnalysisAlgorithmConfigNode(
                name=name,
                algorithm_config_id=algorithm_config_id
            ))

    def get(self) -> List:

        return self.__analysis_config_nodes



#!/usr/bin/env python

import click
import asyncio
import pandas as pd
import yaml
import json
import elements.cli.lib.utils as tsu
from elements.cli.lib.aliased_group import AliasedSuperGroup
import elements.cli.lib.workflow as wf
from elements.cli.commands.algorithm_version import version
from elements.cli.commands.algorithm_computation import computation
from elements.cli.commands.algorithm_config import config


@click.command(cls=AliasedSuperGroup, help="'algorithm' super-group")
@click.pass_context
def algorithm(ctx):
    pass


algorithm.add_command(version)
algorithm.add_command(computation)
algorithm.add_command(config)


@algorithm.command('get', help="Get algorithm info")
@click.pass_context
@click.option('-ia', '--algorithm_id', type=str, default=None)
@click.option('-na', '--algorithm_name', type=str, default=None)
@click.option('-t', '--truncate', type=int, default=36,
              help="Truncate columns to this many chars. Use 0 for no truncation. (a full UUID is 36).")
@click.option('-v', '--verbose', is_flag=True, default=False)
def algorithm_get(ctx, algorithm_id=None, algorithm_name=None, truncate=0, verbose=False):

    wf.check_environment_complete(raise_on_failure=True, verbose=verbose)
    algos = []
    if algorithm_id:
        algos += asyncio.run(wf.get_algorithms(algorithm_ids=[algorithm_id]))
    if algorithm_name:
        algos += asyncio.run(wf.list_algorithms(algorithm_name=algorithm_name))

    data = []
    for algo in algos:
        row = [algo.name, algo.id, algo.author]
        if truncate:
            row = [v[:truncate] for v in row]
        data.append(row)

    df = pd.DataFrame(data=data, columns=['name', 'algorithm_id', 'author'])
    tsu.set_pandas_display()
    click.echo(df)


@algorithm.command('create', help="Create an algorithm object (no manifest)")
@click.pass_context
@click.option('-na', '--algorithm_name', type=str)
@click.option('-a', '--author', type=str, default='TERRASCOPE_AUTHOR')
@click.option('-dn', '--display_name', type=str)
@click.option('-v', '--verbose', is_flag=True, default=False)
@click.option('--dry_run', is_flag=True, default=False)
def algorithm_create(
        ctx,
        algorithm_name,
        author,
        display_name,
        verbose=False,
        dry_run=False,
):

    wf.check_environment_complete(raise_on_failure=True, verbose=verbose)

    author = tsu.get_author(author, raise_on_failure=True)
    click.echo("Using the following information:")
    click.echo(f"--> name:         {algorithm_name}")
    click.echo(f"--> author:       {author}")
    click.echo(f"--> display_name: {display_name}")

    # create the manifest
    if not dry_run:
        algorithm_id = asyncio.run(wf.create_algorithm(
            algorithm_name,
            author,
            display_name,
        ))
        tsu.echo_highlight_suffix("You're new algorithm_id is: ", algorithm_id, 'green')


@algorithm.command('register', help="Create an algorithm, and register it. (with manifest)")
@click.pass_context
@click.option('-m', '--manifest_yaml', type=str)
@click.option('-om', '--output_manifest', type=str, help="Can be json or yaml file")
@click.option('-na', '--algorithm_name', type=str)
@click.option('-a', '--author', type=str, default='TERRASCOPE_AUTHOR')
@click.option('-dn', '--display_name', type=str)
@click.option('-d', '--docker_version_hash', type=str)
@click.option('-pv', '--value_price', default=wf.VALUE_PRICE_DEFAULT, type=float)
@click.option('-pe', '--execution_price', default=wf.EXECUTION_PRICE_DEFAULT, type=float)
@click.option('-nc', '--config_name', default=None, type=str,
              help="Specify this to create an initial algorithm config.")
@click.option('--dry_run', is_flag=True, default=False)
@click.option('-V', '--version', type=str, default=None)
@click.option('-v', '--verbose', is_flag=True, default=False)
def algorithm_register(
        ctx,
        manifest_yaml,
        output_manifest=None,
        algorithm_name=None,
        author=None,
        display_name=None,
        docker_version_hash=None,
        value_price=wf.VALUE_PRICE_DEFAULT,
        execution_price=wf.EXECUTION_PRICE_DEFAULT,
        config_name=None,
        dry_run=False,
        version=None,
        verbose=False,
):

    wf.check_environment_complete(raise_on_failure=True, verbose=verbose)
    # create the manifest
    input_manifest = None
    with open(manifest_yaml, 'r') as fp:
        input_manifest = yaml.safe_load(fp)
    algorithm_name = tsu.override_manifest_param('name', algorithm_name, input_manifest, do_overwrite=True)
    display_name = tsu.override_manifest_param('display_name', display_name, input_manifest, do_overwrite=True)
    author = tsu.get_author(author, raise_on_failure=True)
    # if the author is in the manifest, overwrite it with the command line / env-var value.
    # This is only done in case the user wants to write the manifest to a file,
    # ... we want that file to contain what was actually used.
    if 'author' in input_manifest:
        input_manifest['author'] = author
    new_version = input_manifest['metadata']['version']
    if version:
        new_version = tsu.set_version(version, input_manifest, do_overwrite=True)

    visualizer_config_names = input_manifest.get('visualizer_config_names', None)
    image = wf.update_manifest_docker_hash(input_manifest, docker_version_hash)
    manifest = wf.create_algorithm_manifest(input_manifest)

    click.echo("Using the following information:")
    click.echo(f"--> name:         {algorithm_name}")
    click.echo(f"--> version:      {new_version}")
    click.echo(f"--> author:       {author}")
    click.echo(f"--> display_name: {display_name}")
    click.echo(f"--> image:        {image}")
    click.echo(f"--> prices:       {value_price} / {execution_price}")
    click.echo(f"--> vis-names:    {visualizer_config_names}")

    # create the manifest
    if not dry_run:
        algorithm_id, algorithm_version_id = asyncio.run(wf.new_algorithm(
            algorithm_name,
            author,
            display_name,
            manifest,
            value_price=value_price,
            execution_price=execution_price,
            visualizer_config_names=visualizer_config_names,
        ))
        tsu.echo_highlight_suffix("You're new algorithm_id is: ", algorithm_id, 'green')
        tsu.echo_highlight_suffix("You're new algorithm_version_id is: ", algorithm_version_id, 'green')

        if config_name:
            algorithm_config_id = create_config(algorithm_version_id, config_name, input_manifest)
            tsu.echo_highlight_suffix("You're new algorithm_config_id is: ", algorithm_config_id, 'green')

    if output_manifest:
        with open(output_manifest, 'w') as fp:
            if output_manifest.endswith('json'):
                json.dump(input_manifest, fp, indent=4)
            else:
                yaml.dump(input_manifest, fp)


@algorithm.command('update', help="Update an existing algorithm with manifest or pricing.")
@click.pass_context
@click.option('-m', '--manifest_yaml', type=str, required=True)
@click.option('-ia', '--algorithm_id', type=str, required=True)
@click.option('-om', '--output_manifest', type=str)
@click.option('-d', '--docker_version_hash', type=str, default=None)
@click.option('-pv', '--value_price', default=wf.VALUE_PRICE_DEFAULT, type=float)
@click.option('-pe', '--execution_price', default=wf.EXECUTION_PRICE_DEFAULT, type=float)
@click.option('--dry_run', is_flag=True, default=False)
@click.option('-nc', '--config_name', default=None, type=str,
              help="Specify this to create an initial algorithm config.")
@click.option('-V', '--version', type=str, default=None)
@click.option('-v', '--verbose', is_flag=True, default=False)
def algorithm_update(
        ctx,
        manifest_yaml,
        algorithm_id,
        output_manifest=None,
        docker_version_hash=None,
        value_price=wf.VALUE_PRICE_DEFAULT,
        execution_price=wf.EXECUTION_PRICE_DEFAULT,
        config_name=None,
        dry_run=False,
        version=None,
        verbose=False,
):

    wf.check_environment_complete(raise_on_failure=True, verbose=verbose)
    command_line_version = version

    # create the manifest
    input_manifest = None
    with open(manifest_yaml, 'r') as fp:
        input_manifest = yaml.safe_load(fp)
    algorithm_version = input_manifest['metadata']['version']
    visualizer_config_names = input_manifest.get('visualizer_config_names', None)
    if command_line_version:
        if command_line_version in ['patch', 'minor', 'major']:
            level = command_line_version
            current_algorithm_version = asyncio.run(wf.get_current_algorithm_version(algorithm_id=algorithm_id))
            new_algorithm_version = tsu.increment_version(current_algorithm_version, level=level)
            click.echo(f"The current algorithm version is: {current_algorithm_version}.")
            click.echo(f"Auto-incrementing at {level}-level to: {new_algorithm_version}")
        else:
            new_algorithm_version = command_line_version
        algorithm_version = tsu.set_version(new_algorithm_version, input_manifest, do_overwrite=True)

    image = wf.update_manifest_docker_hash(input_manifest, docker_version_hash)
    manifest = wf.create_algorithm_manifest(input_manifest)

    click.echo("Using the following information:")
    click.echo(f"--> algorithm_id:  {algorithm_id}")
    click.echo(f"--> image:         {image}")
    click.echo(f"--> prices:        {value_price} / {execution_price}")
    click.echo(f"--> version:       {algorithm_version}")
    click.echo(f"--> viz-names:     {visualizer_config_names}")

    # create the manifest
    if not dry_run:
        algorithm_version_id = asyncio.run(wf.update_algorithm(
            algorithm_id,
            manifest,
            value_price=value_price,
            execution_price=execution_price,
            visualizer_config_names=visualizer_config_names,
        ))
        tsu.echo_highlight_suffix("Your updated algorithm_version_id is: ", algorithm_version_id, 'green')

        if config_name:
            algorithm_config_id = create_config(algorithm_version_id, config_name, input_manifest)
            tsu.echo_highlight_suffix("You're new algorithm_config_id is: ", algorithm_config_id, 'green')

    if output_manifest:
        with open(output_manifest, 'w') as fp:
            if output_manifest.endswith('json'):
                json.dump(input_manifest, fp, indent=4)
            else:
                yaml.dump(input_manifest, fp)


def create_config(algorithm_version_id, config_name, input_manifest):

    # use a canned description as this is the first config for this algo version
    config_desc = f"Initial algorithm config for {config_name}"

    # if a config_name is provided, then we'll create an initial config
    # Get the necessary params from the manifest.
    try:
        data_source = input_manifest['inputs'][0]['data_source_name']
    except Exception:
        click.secho("Data source must be provided in manifest", fg='red')
        raise

    try:
        data_type = input_manifest['inputs'][0]['data_type_name']
    except Exception:
        click.secho("Data Type must be provided in manifest", fg='red')
        raise

    image_processing_spec = None
    if 'parameters' in input_manifest['inputs'][0]:
        image_processing_spec = input_manifest['inputs'][0]['parameters'].get('image_processing_spec', None)
    data_parameters = {}
    if image_processing_spec:
        data_parameters['image_processing_spec'] = image_processing_spec

    algorithm_config_id = asyncio.run(
        wf.create_algorithm_config(
            algorithm_version_id,
            config_name,
            config_desc,
            tsu.FakeDataSource(data_source),
            data_type,
            data_parameters,
        )
    )
    return algorithm_config_id



import logging
from enum import Enum
from typing import List, Dict

from google.protobuf.json_format import MessageToDict
from google.protobuf.struct_pb2 import Struct

from elements.sdk.builder.toi import Frequency


class DataType(Enum):
    PINGS = "pings"
    DEVICE_TRACKS = "device_tracks"
    DEVICE_VISITS = "device_visits"
    OPTICAL_IMAGERY = "optical_imagery"
    OBJECT_DETECTION = "object_detection"
    MULTICLASS_OBJECT_DETECTION = "multiclass_object_detection"
    SPOT_MOSAIC = "spot_mosaic"
    PSSCENE = "psscene"
    SKYSAT = "skysat"
    POPULATION = "population"
    PROXIMITY_ZONE_UDC = "proximity_zone_udc"
    HEURISTIC_JUMPS = "heuristic_jumps"
    NAVIGATIONAL_QUALITY_INDICATORS = "navigational_quality_indicators"
    GNSS_CLUSTERS = "gnss_clusters"
    CROP_CIRCLES = "crop_circles"
    TRACEABILITY_CLUSTERS = "traceability_clusters"
    TRACEABILITY_HEATMAP = "traceability_heatmap"
    UDC_HEATMAP = "unique_device_count_heatmap"
    UDC = "unique_device_count"
    AIRBUS_PLEIADES = "airbus_pleiades"
    PLANET_PSSCENE4BAND = "planet_PSScene4Band"


class DataSource(Enum):
    # Ping Sources
    GEOLOCATION = "geolocation_pings"
    SAFEGRAPH = "safegraph_pings"
    XMODE = "xmode_pings"
    CUEBIQ = "cuebiq_pings"
    BLOGWATCHER = "blogwatcher_pings"
    SPIRE = "spire_pings"
    SPIRE_NMEA = "spire-nmea_pings"
    CUEBIQ_WORKBENCH = "cuebiq-workbench_pings"
    WEJO = "wejo_pings"
    HAWKEYE360 = "hawkeye360_pings"
    OTONOMO = "otonomo_pings"
    ADSBX = "adsbx_pings"
    EXACTEARTH = "exact-earth_pings"
    EXACTEARTH_NMEA = "exact-earth_nmea-pings"

    # Imagery Sources
    IMAGERY = "imagery_imagery"
    PLANET_REORTHOTILE = "planet_REOrthoTile"
    PLANET_PSORTHOTILE = "planet_OSOrthoTile"
    PLANET_PSSCENE4BAND = "planet_PSScene4Band"
    # PLANET_PSSCENE4BAND = "PL-PSScene4Band"
    PLANET_SKYSATSCENE = "planet_SkySatScene"
    PLANET_SKYSATCOLLECT = "planet_SkySatCollect"
    AIRBUS_SPOT7 = "airbus_SPOT7"
    AIRBUS_PHR1A = "airbus_PHR1A"
    AIRBUS_PHR1B = "airbus_PHR1B"
    AIRBUS_PLEIADES = "airbus_pleiades"
    MAXAR_WORLDVIEW_2 = "maxar_Worldview_02"
    AIRBUS_SPOT = "airbus_SPOT"


class ReleaseStatus(Enum):
    DEV = "DEV"
    QA = "QA"
    STAGE = "STAGE"
    PROD = "PROD"


class InterfaceType(Enum):
    SYNCHRONOUS_TASK_HTTP_WORKER = "SYNCHRONOUS_TASK_HTTP_WORKER"
    FILESYSTEM_TASK_WORKER = "FILESYSTEM_TASK_WORKER"
    NO_OP_WORKER = "NO_OP_WORKER"


class ParallelType(Enum):
    AOI = "AOI"
    TIME_RANGE = "TIME_RANGE"


class PermissionType(Enum):
    ALLOWED = 'Allowed'
    DISALLOWED = 'Disallowed'


class AlgorithmManifest:
    """
    This is a basic builder to help build a factories in a safe way instead of managing json configuration
    files. This includes all the methods to incrementally build out an algorithm manifest proto struct that can be
    used for an algorithm.

    REQUIRED Methods:
    The following methods have to be run in sequence. I don't have a better way to do this yet so for now we will just
    call these methods. Each time a new algorithm manifest is need use the following sequence.
    - metadata_required()
    - interface_required()
    - inputs_add_data_type(): can call this more than once, need AT LEAST 1
    - container_parameters_required()
    - outputs_add_data_type(): can call more than once, need AT LEAST 1

    """

    def __init__(self):
        self.__metadata = {
            "metadata": {}
        }
        self.__interface = {
            "interface": {}
        }
        self.__inputs = []
        self.container_parameters = {
            "container_parameters": {}
        }
        self.__parameters = {
            "parameters": []
        }
        self.__grouping = {
            "grouping": {}
        }
        self.__parallelization = {
            "parallelization": []
        }
        self.__restrictions = {
            "restrictions": {}
        }
        self.__outputs = []
        self.__performance_metrics = {
            "performance_metrics": {}
        }
        self.__manifest_version = {
            "manifest_version": "0.1.0"
        }
        self.__required = {
            "metadata": {
                "description": False,
            },
            "interface": {
                "interface_type": False
            },
            "inputs": False,
            "container_parameters": {
                "image": False,
                "command": False
            },
            "outputs": False,
            "manifest_version": True,
        }

    # Metadata
    def metadata_required(self, description: str):
        self.metadata_description(description=description)

    def manifest_version(self, manifest_version: str):
        self.__manifest_version['manifest_version'] = manifest_version
        self.__required['manifest_version'] = True

    def metadata_description(self, description: str):
        self.__metadata['metadata']['description'] = description
        self.__required['metadata']['description'] = True

    def metadata_indicator(self, indicator: str):
        self.__metadata['metadata']['indicator'] = indicator

    def metadata_tags(self, tags: List):
        self.__metadata['metadata']['tags'] = tags

    # Interface
    def interface_required(self, interface_type: str):
        self.interface_interface_type(interface_type=interface_type)

    def interface_interface_type(self, interface_type: str):
        self.__interface['interface']["interface_type"] = interface_type
        self.__required['interface']['interface_type'] = True

    def interface_adapter(self, adapter: str):
        self.__interface['interface']['adapter'] = adapter

    # Inputs
    def inputs_add_data_type(self, data_type_name: str, **kwargs):
        """
        [REQUIRED]

        Adds data types this algorithm is allowed to have as inputs.

        :param data_type_name:
        :param kwargs:
        :return:
        """
        entry = {
            "data_type_name": data_type_name
        }
        for key in ['min_count', 'max_count', 'parameters', 'bands', 'data_source_ids']:
            if key in kwargs.keys():
                entry[key] = kwargs[key]

        self.__inputs.append(entry.copy())
        self.__required['inputs'] = True

    # Container Parameters
    def container_parameters_required(self, image: str, command: List):
        self.container_parameters_image(image=image)
        self.container_parameters_command(command=command)

    def container_parameters_image(self, image: str):
        self.container_parameters['container_parameters']['image'] = image
        self.__required['container_parameters']['image'] = True

    def container_parameters_command(self, command: List):
        self.container_parameters['container_parameters']['command'] = command
        self.__required['container_parameters']['command'] = True

    def container_parameters_resource_request(self, **kwargs):
        if 'resource_request' not in self.container_parameters['container_parameters'].keys():
            self.container_parameters['container_parameters']['resource_request'] = {}

        for resource_request_key in ['gpu', 'memory_gb', 'cpu_millicore', 'max_input_gb']:
            if resource_request_key in kwargs.keys():
                self.container_parameters['container_parameters']['resource_request'][resource_request_key] = kwargs[resource_request_key]

    # Parameters
    def parameter_add(self, **kwargs):
        assert 'description' in kwargs.keys()
        entry = {}
        if 'name' in kwargs.keys():
            entry['name'] = kwargs['name']
        if 'type' in kwargs.keys():
            entry['type'] = kwargs['type']
        if 'unit' in kwargs.keys():
            entry['unit'] = kwargs['unit']
        if 'description' in kwargs.keys():
            entry['description'] = kwargs['description']
        if 'min' in kwargs.keys():
            entry['min'] = kwargs['min']
        if 'max' in kwargs.keys():
            entry['max'] = kwargs['max']
        if 'allowed_values' in kwargs.keys():
            entry['allowed_values'] = kwargs['allowed_values']
        if 'default' in kwargs.keys():
            entry['default'] = kwargs['default']

        self.__parameters['parameters'].append(entry.copy())

    # Grouping
    def grouping_frequency(self, frequency: str, value: int):
        assert isinstance(value, int)
        assert isinstance(frequency, str)
        self.__grouping['grouping']['frequency'] = frequency
        self.__grouping['grouping']['value'] = value

    # Parallelization
    def parallelization_add_configuration(self, config: Dict):
        self.__parallelization['parallelization'].append(config)

    # Restrictions
    def restriction_spatial(self, permission_type: PermissionType,
                            overridable: bool,
                            geometry):
        self.__restrictions['spatial_restriction']['permission_type'] = permission_type
        self.__restrictions['spatial_restriction']['overridable'] = overridable
        self.__restrictions['spatial_restriction']['geometry'] = geometry

    def restriction_temporal_restriction(self, tois: List):
        self.__restrictions['temporal_restriction'] = tois

    def restriction_size(self):
        # TODO
        pass

    def outputs_add_data_type(self, data_type_name: str, **kwargs):
        """
        Adds data types this algorithm is allowed to have as outputs.
        """
        entry = {
            "data_type_name": data_type_name
        }
        if 'observation_value_columns' in kwargs.keys():
            entry['observation_value_columns'] = kwargs['observation_value_columns']
        if 'sample_result' in kwargs.keys():
            entry['sample_result'] = kwargs['sample_result']
        if 'classes' in kwargs.keys():
            entry['classes'] = kwargs['classes']
        if 'output_geometry' in kwargs.keys():
            entry['output_geometry'] = kwargs['output_geometry']
        if 'skip_export' in kwargs.keys():
            entry['skip_export'] = kwargs['skip_export']

        self.__outputs.append(entry.copy())
        self.__required['outputs'] = True

    # Performance Metrics

    def get(self) -> Struct:
        """
        Description:

        Once your manifest is built, use this method to pass the manifest into your TerraScope API Calls.
        This will construct the proper format for that, and check that all the required fields are present.

        :return: An Algorithm Manifest
        """
        manifest = Struct()
        manifest.update(self.__metadata)
        manifest.update(self.__interface)
        manifest.update({"inputs": self.__inputs})
        manifest.update(self.container_parameters)
        manifest.update({"outputs": self.__outputs})
        manifest.update(self.__manifest_version)

        if len(self.__performance_metrics) > 0:
            manifest.update(self.__parameters)
        if len(self.__grouping.keys()) > 0:
            manifest.update(self.__grouping)
        if self.__parallelization['parallelization']:
            manifest.update(self.__parallelization)
        if len(self.__restrictions.keys()) > 0:
            manifest.update(self.__restrictions)
        if len(self.__performance_metrics.keys()) > 0:
            manifest.update(self.__performance_metrics)

        for outer_key in self.__required.keys():
            if isinstance(outer_key, Dict):
                for inner_key in outer_key.keys():
                    if not self.__required[outer_key][inner_key]:
                        logging.info("[{} : {}] - was not set, and is required.".format(outer_key, inner_key))
                        assert self.__required[outer_key][inner_key]
            else:
                if not self.__required[outer_key]:
                    logging.info("[{}] - was not set, and is required.".format(outer_key))
                    assert self.__required[outer_key]

        return manifest


class AlgorithmConfiguration:

    def __init__(self):
        """
        This is an Algorithm Configuration Factory. The purpose is to simplify building an algo configuration struct.
        When creating an algorithm config, specifying a data source is required. Optionally, you may also specify
        modifications to algorithm parameters that were set in the algorithm manifest.

        Defaults:
            - Grouping Frequency: DAILY, 1
        """
        self.__required = {
            "data_sources": False,
            "grouping": False
        }
        self.__algorithm_parameters = {
            "parameters": {}
        }
        self.__grouping = {
            "grouping": {}
        }
        self.__data_sources = []
        self.__outputs = []
        # Any additional configuration
        self.__additional_configuration = {}
        # Set Default grouping frequency
        self.grouping_frequency(Frequency.DAILY, 1)

    def add_data_source(self, data_type: str, **kwargs):

        data_source = kwargs.pop('data_source', None)
        data_sources = kwargs.pop('data_sources', [])
        if data_source and data_sources:
            logging.error("Only specify either data_source or data_sources for add_data_source.")
            return
        if data_source is not None:
            data_sources.append(data_source)

        for source in self.__data_sources:
            # Check if Data Type and Data Source combination Exists
            if source['data_type_name'] == data_type:
                for source_id in source['data_source_ids']:
                    for data_source in data_sources:
                        if source_id == data_source.value:
                            logging.info("data source [{}] already registered.".format(data_source))
                            return

        data_parameters = kwargs.pop('data_parameters', None)
        new_data_source = {
            "data_type_name": data_type,
            "data_source_ids": [data_source.value for data_source in data_sources]
        }
        new_data_source.update(kwargs)
        if data_parameters is not None:
            new_data_source.update({
                "parameters": data_parameters
            })
        self.__data_sources.append(new_data_source)

        self.__required['data_sources'] = True

    def update_output_data_type(self, data_type: str, skip_export: bool = False,
                                observation_value_columns: List[str] = None, data_transfer_params: Dict = None):
        data_type_updated = False
        observation_value_columns = observation_value_columns if observation_value_columns else []
        data_transfer_params = data_transfer_params if data_transfer_params else {}
        for source in self.__outputs:
            if source['data_type_name'] == data_type:
                data_type_updated = True
                source['skip_export'] = skip_export
                source['observation_value_columns'] = observation_value_columns
                source['data_transfer'] = data_transfer_params

        if not data_type_updated:
            self.__outputs.append({
                'data_type_name': data_type,
                'skip_export': skip_export,
                'observation_value_columns': observation_value_columns,
                'data_transfer': data_transfer_params
            })

    def add_algorithm_parameter(self, key: str, value):
        for param in self.__algorithm_parameters['parameters'].keys():
            if key == param:
                assert key != param

        self.__algorithm_parameters['parameters'][key] = value

    def update_configuration(self, configuration: Dict):
        self.__additional_configuration.update(configuration)

    # Grouping
    def grouping_frequency(self, frequency: Frequency, value: int):
        assert isinstance(frequency, Frequency)
        assert isinstance(value, int)
        self.__grouping['grouping']['frequency'] = Frequency(frequency).name
        self.__grouping['grouping']['value'] = value
        self.__required['grouping'] = True

    def get(self) -> Struct:
        """
        Description:

        Once your manifest is built, use this method to pass the manifest into your TerraScope API Calls.
        This will construct the proper format for that, and check that all the required fields are present.

        :return: An Algorithm Configuration as struct_pb2.Struct
        """

        for key in self.__required.keys():
            if not self.__required[key]:
                logging.info("[{}] - was not set, and is required.".format(key))
                assert self.__required[key]

        algorithm_config = Struct()
        algorithm_config.update({"data_sources": self.__data_sources})
        algorithm_config.update({"outputs": self.__outputs})
        algorithm_config.update(self.__algorithm_parameters)
        algorithm_config.update(self.__grouping)
        algorithm_config.update(self.__additional_configuration)

        MessageToDict(algorithm_config)
        return algorithm_config
